////////////////////////////////////////////////////////////////////////
//
//      Quinn Jacobson (qjacobso@ece.wisc.edu)
//      ECE department U of Wisconsin
//
//      CS 701
//      Feb 1996
//
//      LocalVars.cc
//
////////////////////////////////////////////////////////////////////////
//
// Code to maintain information about local variables.
//
////////////////////////////////////////////////////////////////////////

//----- Defins and Includes -----
 
#define	LVsize	1024

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "LocalVars.h"

int LVpOffset;
int LVvOffset;

//----- STRUCTURES -----

typedef struct LVnode 
{
    const char		*name;
    int			offset;
    int			size;
    struct LVnode	*next;
} LVnode;

//----- Globals -----

static LVnode	*LVtable[LVsize];


//----- CODE -----

//******************** 
int hash(const char *s)
// Simple Hash function
{
    int r;
    int i;
    char c;
    
    r = 0;
    for(i = 0; i < 8; i++)
    {
	if ((c = s[i]) == '\0')	
	    break;
	r = (r + c * ((i+1) & 3)) % LVsize;
    }
    //printf("hash() : %s = %d\n", s, r);
    return r;
}

//******************** 
int LVdata_size(char *name)
// Returns the size of local variable
{
    LVnode *p;
    int	    i;

    i = hash(name);
    for (p = LVtable[i]; p; p = p->next)
    {
	if (strcmp(p->name, name) == 0)
	    return (p->size);
    }
    return (0);
}

//******************** 
int LVoffset(char *name)
// Returns offset for a local variable/incoming parameter
//      0 if not a local variable or parameter
//      negative if local variable (relative to %fp)
//      positive if parameter (relative to %fp)
{
    LVnode *p;
    int	    i;

    i = hash(name);
    for (p = LVtable[i]; p; p = p->next)
    {
	if (strcmp(p->name, name) == 0)
	    return (p->offset);
    }
    return (0);
}

//******************** 
int LVpresent(int i, const char *s)
// return 1 if string is present in the table, o otherwise
// i is the hash of s, s is the string being looked for
{
    LVnode *p;

    for (p = LVtable[i]; p; p = p->next)
    {
	if (strcmp(p->name, s) == 0)
	    return (1);
    }
    return (0);
}

//******************** 
void LVparam(const char *name, int size)
// enter an incoming parameter 
{
    int    i;
    LVnode *p;

    //printf("LVparam() : %s, %d, %d\n", name, size, LVpOffset);
    i = hash(name);
    
    p = (LVnode *)malloc(sizeof(LVnode));
    p->name = name;
    p->offset = LVpOffset; 
    p->size = size;
    LVpOffset += size;
    if (LVpOffset % 4)
    	LVpOffset += 4 - (LVpOffset %4);
    p->next = LVtable[i];
    LVtable[i] = p;

    //printf("LVparam() : %s, %d, %d\n", name, size, LVpOffset);
    return;
}


//******************** 
void LVvar(const char *name, int size)
// enter a local variable 
// if the same name has not already been entered as a incoming parameter
{
    int    i;
    LVnode *p;
    int	   lalig;
    int    pad_size;

    if (size == 0) return;
    //printf("LVvar() : %s, %d, %d\n", name, size, LVvOffset);
    i = hash(name);

    if (LVpresent(i, name)) 
    {
      	return;
    }
    if ((size >= 4) && (size < 8))
	lalig = 4;
    else if (size >= 8)
	lalig = 8;
    else lalig = size;
    assert (lalig);
    if (size % lalig)
	pad_size = size + (lalig - (size % lalig));
    else
	pad_size = size;
    assert (pad_size > 0);
    assert (pad_size >= size);
    assert ((pad_size % lalig) == 0);
    p = (LVnode *)malloc(sizeof(LVnode));
    p->name = name;
    p->size = size;
    LVvOffset -= pad_size;
    assert(LVvOffset < -16);
    if ((0-LVvOffset) % lalig) 
    {
	assert(((0-LVvOffset) % lalig) > 0);
        LVvOffset -= (lalig - ((0-LVvOffset) % lalig)); // align data
    }
    p->offset = LVvOffset; 
    p->next = LVtable[i];
    LVtable[i] = p;
    //printf("!LVvar() new: %s, %d, %d\n", name, size, LVvOffset);
}

//******************** 
void LVreset()
// Reset Local Variable data structures 
{
    int i;
    LVnode *p;
    LVnode *n;

    for(i = 0; i < LVsize; i++)
    {
	p = LVtable[i];
	n = NULL;
	while (p)
	{
	    if (n) free(n);
	    n = p;
	    p = p->next;
	}
	if (n) free(n);
        LVtable[i] = NULL; 
    }
    LVvOffset = -272;
    LVpOffset = 68;
}

//******************** 
int LVvarSize()
// return the size of the local variable space on the stack
{
    return(0-LVvOffset);
}
