////////////////////////////////////////////////////////////////////////
//
//      Quinn Jacobson (qjacobso@ece.wisc.edu)
//      ECE department U of Wisconsin
//
//      CS 701
//      Feb/March 1996
//
//      codegen.cc
//
////////////////////////////////////////////////////////////////////////
//
// Top level module for code-generator for Simple-SUIF for the Sparc.
//
// HANDLES:
//    default function header stuff
//    determining frame stack size
//    allocate memory for functions that return records
//    CALLS : reg_alloc() to allocate registers
//    save params that are not only used in registers
//    CALLS : proc_instr() to create code for function 
//
////////////////////////////////////////////////////////////////////////
 

/************************************************************************
 * header files.
************************************************************************/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

extern "C" {
#include <simple.h>
}

#include "codegen.h"
#include "ss_symtab.h"
#include "reg_alloc.h"
#include "LocalVars.h"
#include "proc_instr.h"
#include "temp.h"
#include "instr.h"
#include "reg_class.h"
#include "reg_pass.h"
#include "print.h"

/************************************************************************
 * Globals
************************************************************************/

// ARGs
extern int ARG_post;
extern bool alloc;
extern bool reportAlloc;

extern s_instr *instr_head;
extern s_instr *instr_tail;

extern ss_symbol* 	ss_global_syms;
extern unsigned int  	ss_global_nsyms;
extern char		map_reg[MAX_REG];

unsigned int		ret_record_offset[256];
unsigned int		ret_record_index;
unsigned int		fp_save_offset;

reg_class inreg;
reg_class ccreg;
reg_class fpreg;

/************************************************************************
 * void codegen_init()
************************************************************************/

void
codegen_init() {
    //unsigned int	i;
    //ss_symbol	*s;

// Initialize registers 

    TEMPinit();

    // INIT integer regs
    // %r1 is used for as a scratch register
    // %r2 is used for a work register for operand 1
    // %r3 is used for a work register for operand 2
    // %r4 is used for a work register to hold addresses
    // %r5 - %r8 are used by the OS
    inreg.REGadd(9); inreg.REGadd(10); inreg.REGadd(11); 
    inreg.REGadd(12); inreg.REGadd(13); inreg.REGadd(15);

    // INIT fp regs
    // %f0 is reserved for return value
    // %f30 is reserved for scratch
    // %f2 and %f4 are reserved for work registers
    fpreg.REGadd(6); fpreg.REGadd(8); fpreg.REGadd(10); fpreg.REGadd(12); 
    fpreg.REGadd(14); fpreg.REGadd(16); fpreg.REGadd(18); fpreg.REGadd(20);
    fpreg.REGadd(22); fpreg.REGadd(24); fpreg.REGadd(26); fpreg.REGadd(28);

}




/************************************************************************
 * void codegen_finish()
************************************************************************/

void
codegen_finish() {
}



/************************************************************************
 * size_of_stack()
************************************************************************/

void local_and_param(char* proc_name)
// 1) Reset Param and local variable symbol tables
// 2) determine incoming parameters and enter them in the symbol table
// 3) determine local variables and enter them in the symbol table
//    and allocate space on the stack for them
{
  const ss_symbol   *Lsyms;
  const ss_symbol   *Psyms;
  const ss_symbol   *s;
  unsigned int 	    Lnumsyms;
  unsigned int 	    Pnumsyms;
  unsigned int	    i;

// Reset Locals
    LVreset();

// Create table of parameters and local variables
// Assign them offsets on the stack
    //printf("Enter size_of_stack\n"); fflush(stdout);
    Pnumsyms = ss_proc_nparams(proc_name);
    Psyms = ss_proc_params(proc_name);
    for (i = 0; i < Pnumsyms; i++) 
    {
	s = &(Psyms[i]);
	LVparam(s->sym_name, s->sym_size>>3);
    }
    Lnumsyms = ss_proc_nsyms(proc_name);
    Lsyms = ss_proc_syms(proc_name);
    for (i = 0; i < Lnumsyms; i++) 
    {
	s = &(Lsyms[i]);
        if (s->is_defined == 0)
	    LVvar(s->sym_name, s->sym_size>>3);
    }

}

int size_of_stack(simple_instr* inlist)
// Does many things
// 4) determin the space required for outgoing params
// 5) allocate  space for functions that return structures
// 6) allocate space for storing fp registers accross calls
// 7) determine the overall size of the stack
//
{
    unsigned int	i;
    unsigned int	size_local;	// Size of scratch and local on stack
    unsigned int	size_param;	// Size of outgoing params on stack
    simple_instr	*inst;
    unsigned int	num;		// Number of outgoing params
    unsigned int	temp, rsize;
    unsigned int	size;		// Size of stack frame
    simple_reg		*arg;
    int			param_space;
    int			rr_space;
    int			rr_index;

// Determine largest outgoing param space
    size_param = 24;
    num = 0;
    for (inst = inlist; inst; inst = inst->next)
    {
	if (inst->opcode == CALL_OP)
	{
	// outgoing param space
	    num = inst->u.call.nargs;
	    temp = 0;
	    if (num > 0) arg = *(inst->u.call.args);
	    for (i = 0; i < num; i++)
	    {
	 	if (arg->var)
		{
		    if (arg->var->type->base == RECORD_TYPE)
		    {
			rsize = arg->var->type->len >> 3;
			if (rsize % 4)
			   rsize += 4 - (rsize % 4);
			temp += rsize;
		    }
		    else if (arg->var->type->base == FLOAT_TYPE)
			temp += 8;
		    else
			temp += 4;
		}
		else
		    temp += 4;
	    }
	    if (temp > size_param)
		size_param = temp;
	}
    }
    param_space = size_param;

// Determine stack space for scratch and local variables
    size_local = LVvarSize(); // Already calculated while creating table
    if (size_local % 8)
	size_local += 8 - (size_local % 8);

// return recored space
    ret_record_index = 0;
    rr_index = 0;

    for (inst = inlist; inst; inst = inst->next)
    {
	if (inst->opcode == CALL_OP)
	{
	    if (inst->type->base == RECORD_TYPE);
	    if (inst->u.call.dst)
	    {
		rr_space = inst->u.call.dst->var->type->len >> 3;
		if (rr_space % 8)
		    rr_space += 8 - (rr_space % 8);
		size_local += rr_space;
		ret_record_offset[rr_index] = 0 - size_local;
		rr_index++;
		assert (rr_index < 256);
	    }
	}
    }

// Offsets
    fp_save_offset = 64 + param_space + 4;
    if (fp_save_offset % 8)
       fp_save_offset += 8 - (fp_save_offset % 8);

// Size of stack frame
    size = size_local + fp_save_offset + 96;
    assert((size % 8) == 0);
    if (size % 8)
	size += 8 - (size % 8);

    return size;
}

void save_params(char* proc_name)
// Traverse the incoming parameter list and store any that are in registers
// back to memory if they are not globally allocated for live range including
// zero block.
// Globally allocated params that are in the zero block and are not passed 
// in registers must be loaded into registers.
{
    const ss_symbol	*Psyms;
    const ss_symbol	*s;
    unsigned int 	Pnumsyms;
    unsigned int	i, r;
    int 		dat_size, trn_size;

    Pnumsyms = ss_proc_nparams(proc_name);
    Psyms = ss_proc_params(proc_name);
    r = 0;
    for (i = 0; i < Pnumsyms; i++) 
    {
	s = &(Psyms[i]);
	if (r < 6)
	{
	    // STORE IF NECESSARY
	    dat_size = s->sym_size >> 3;

/////////////////////////////////////////////////////////////////////////
// CS701 PROGRAM #2:
// If parameter is allocated via global register allocation:
//     1> move between registers if required
//     2> continue
/////////////////////////////////////////////////////////////////////////

	    // STORE REQUIRED
	    for (trn_size=0; (trn_size < dat_size)&&(r < 6); trn_size += 4)
	    {
	      // BUG FIX: Don't always generate a "st" instruction to
	      //          store the parameter.  Generate the appropriate
	      //          instruction based on the size of the value.
	      switch (s->sym_size) {

	      case 8:
	        printf("\tstb %%i%d,[%%fp+%d]\n", (r),(68 + (4*r)));
		break;

	      case 16:
	        printf("\tsth %%i%d,[%%fp+%d]\n", (r),(68 + (4*r)));
		break;

	      default:
	        printf("\tst %%i%d,[%%fp+%d]\n", (r),(68 + (4*r)));
	      } // end switch (s->sym_size)

	      r += 1;
	    }
	} 
	else // is a parameter past 6
	{

/////////////////////////////////////////////////////////////////////////
// CS701 PROGRAM #2:
// If parameter is allocated via global register allocation from a range
// that is live in the begining of the function load it from memory
/////////////////////////////////////////////////////////////////////////

	}
    }
    return;
}

/************************************************************************
 * simple_instr* do_procedure(simple_instr* inlist, char* proc_name)
************************************************************************/

simple_instr*
do_procedure(simple_instr* inlist, char* proc_name) 
{
    unsigned int	size_frame;
    int			i;

    basic_block_node* cfg; // the head of the cfg

// standard procedure header
    printf("\t.text\n");
    printf("\t.align 4\n");
    // printf("\t.global _%s\n", proc_name);
    printf("\t.global %s\n", proc_name);
    printf("\t.proc 04\n");
    // printf("_%s:\n", proc_name);
    printf("%s:\n", proc_name);
    fflush(stdout);

// Locals and Params
    local_and_param(proc_name);
   
/////////////////////////////////////////////////////////////////////////
// CS701 PROGRAM #2:
// Build CFG and do register allocation here
/////////////////////////////////////////////////////////////////////////
    //debug_instr_types(inlist);
    if (alloc) {
        cfg = build_cfg(inlist);
        if (reportAlloc) {
            print_cfg(cfg);
        }
        delete_cfg(cfg);
    }

// Stack pointer advancement
    size_frame = size_of_stack(inlist);
    printf("\tsave %%sp,-%d,%%sp\n", size_frame);
    fflush (stdout);

    reg_pass(inlist);

// Allocate remaining local registers for temp registers
    ccreg.REGinit(0, -80);

    for (i = 16; i <= 29;  i++)
    {
        /////////////////////////////////////////////////////////////////
        // CS701 PROGRAM #2:
        // MUST MODIFY THIS
        // Only make this call if register %ri is not used in global
	// register allocation
        /////////////////////////////////////////////////////////////////
	    ccreg.REGadd(i);
    }

// Save params that need to be saved
    save_params(proc_name);

// Special main case
    if (strcmp(proc_name, "main") == 0)
    {
	// SOLARIS: printf("\tcall ___main,0\n");
	// SOLARIS: printf("\tnop\n");
    }
    
    printf("! Start instruction selection\n"); fflush(stdout);
// Process Instructions
    instr_init();
    proc_instructions(inlist, proc_name);	// process suif instructions
						// and generate sparc instrns
    printf("! instruction selection OK\n"); fflush(stdout);

/////////////////////////////////////////////////////////////////////////
// CS701 PROGRAM #2:
// Do post pass stuff here on SPARC instruction list instr_head if needed
/////////////////////////////////////////////////////////////////////////

    print_instr_list(instr_head);

// Standard ending
    printf("_END_%s:\n", proc_name); 

// Clean UP
    return inlist;
}
