////////////////////////////////////////////////////////////////////////
//
//      Quinn Jacobson (qjacobso@ece.wisc.edu)
//      ECE department U of Wisconsin
//
//      CS 701
//      April 1996
//
//      instr.h
//
////////////////////////////////////////////////////////////////////////
//
// Defination of instr type
//
////////////////////////////////////////////////////////////////////?????

//------------ Includes --------
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "instr.h"

//----------- STRUCTS -----------
typedef struct label_linker
{
    char 		*label;
    struct s_instr	*instr;
    struct label_linker	*next;
} label_linker;

//----------- GLOBALS -----------
s_instr	*instr_head = NULL;
s_instr	*instr_tail = NULL;

label_linker *labels = NULL;


//----------- CODE -----------

void instr_init()
{
    s_instr *ptr;
    s_instr *temp;
    label_linker *lab;
    label_linker *temp_lab;

    ptr = instr_head;
    while (ptr)
    {
	temp = ptr;
	ptr = ptr->next;
	if (temp->src1) free(temp->src1);
	if (temp->src2) free(temp->src2);
	if (temp->dst) free(temp->dst);
	free(temp);
    }
    instr_head = NULL;
    instr_tail = NULL;

    lab = labels;
    while (lab)
    {
	temp_lab = lab;
	lab = lab->next;
	free(temp_lab);
    }
    labels = NULL;
}

//**********
s_instr* new_instr(int opcode)
{
    s_instr	*instr;
    instr = (s_instr*)malloc(sizeof(s_instr));
    assert (instr);
    instr->opcode = opcode;
    instr->src1 = NULL;
    instr->src2 = NULL;
    instr->dst = NULL;
    instr->target = NULL;
    if (instr_tail)
    {
	instr->prev = instr_tail;
	instr->next = NULL;
	instr_tail->next = instr;
	instr_tail = instr;
    }
    else
    {
	instr_head = instr;
	instr_tail = instr;
	instr->next = NULL;
	instr->prev = NULL;
    }
    return (instr);
}

//**********
s_operand *new_operand()
{
    s_operand *operand;
    operand = (s_operand *)malloc(sizeof(s_operand));
    assert (operand);
    operand->type = -1;
    operand->reg_num = -1;
    operand->immed = 0;
    operand->string = NULL;
    return (operand);
}

//**********
s_operand *zero_operand()
{
    s_operand *operand;
    operand = new_operand();
    operand->type = 1;
    operand->reg_num = 0;
    return (operand);
}

//**********
s_operand *reg_operand(int reg_num)
{
    assert((reg_num >= 0) && (reg_num <= 63));
    s_operand *operand;
    operand = new_operand();
    operand->type = 1;
    operand->reg_num = reg_num;
    return (operand);
}

//**********
s_operand *jump_target(char *target, char* proc_name)
{
    s_operand *operand;
    char *str;
    int len;

    len = strlen(target) + strlen(proc_name) + 3;
    str = (char *)malloc(len * sizeof(char));
    strcpy(str, target);
    // SOLARIS: strcat(str, "_");
    strcat(str, proc_name);
    operand = new_operand();
    operand->type = 0;
    operand->string = str;
    return (operand);

}

//**********
s_operand *quinn_label(char *target, char* quinn_num)
{
    s_operand *operand;
    char *str;
    int len;

    len = strlen(target) + strlen(quinn_num) + 4;
    str = (char *)malloc(len * sizeof(char));
    strcpy(str, target);
    // SOLARIS: strcat(str, "_");
    strcat(str, quinn_num);
    strcat(str, ":");
    operand = new_operand();
    operand->type = 0;
    operand->string = str;
    return (operand);

}

//**********
s_operand *fp_offset_operand(int offset)
{
    s_operand *operand;

    operand = new_operand();
    operand->type = 9;
    operand->reg_num = 30;
    operand->immed = offset;
    return (operand);
}

//**********
s_operand *sp_offset_operand(int offset)
{
    s_operand *operand;

    operand = new_operand();
    operand->type = 9;
    operand->reg_num = 14;
    operand->immed = offset;
    return (operand);
}

//**********
s_operand *immed_operand(int immed)
{
    s_operand *operand;

    operand = new_operand();
    operand->type = 2;
    operand->immed = immed;
    return (operand);
}

//**********
s_operand *simple_address(int reg, int offset)
{
    s_operand *operand;

    assert((reg >= 0) && (reg <= 63));
    operand = new_operand();
    operand->type = 9;
    operand->reg_num = reg;
    operand->immed = offset;
    return (operand);
}

//**********
s_operand *hi_operand(char *label, int immed)
{
    s_operand *operand;
    char *str;
    int len;

    operand = new_operand();
    operand->immed = immed;
    if (label)
    {
    	len = strlen(label) + 3;
    	str = (char *)malloc(len * sizeof(char));
    	// SOLARIS: strcpy(str, "_");
    	strcpy(str, label);
	operand->string = str;
	if (immed)
	{
	    operand->type = 8;
	}
	else
	{
	    operand->type = 6;
	}
    }
    else
	operand->type = 7;

    return (operand);
}

//**********
s_operand *lo_operand(char *label, int immed)
{
    s_operand *operand;
    char *str;
    int len;

    operand = new_operand();
    operand->immed = immed;
    if (label)
    {
    	len = strlen(label) + 3;
    	str = (char *)malloc(len * sizeof(char));
    	// SOLARIS: strcpy(str, "_");
    	strcpy(str, label);
	operand->string = str;
	if (immed)
	{
	    operand->type = 5;
	}
	else
	{
	    operand->type = 3;
	}
    }
    else
	operand->type = 4;

    return (operand);
}

//**********
s_operand *call_target(char *label)
{
    s_operand *operand;
    char *str;
    int len;

    len = strlen(label) + 3;
    str = (char *)malloc(len * sizeof(char));
    assert(str);
    // SOLARIS: strcpy(str, "_");
    strcpy(str, label);

    operand = new_operand();
    operand->type = 0;
    operand->string = str;
    return (operand);
}

//**********
s_operand *label_string(char *target, char* proc_name)
{
    s_operand *operand;
    char *str;
    int len;

    len = strlen(target) + strlen(proc_name) + 4;
    str = (char *)malloc(len * sizeof(char));
    strcpy(str, target);
    // SOLARIS: strcat(str, "_");
    strcat(str, proc_name);
    strcat(str, ":");
    operand = new_operand();
    operand->type = 0;
    operand->string = str;
    return (operand);

}

//**********
s_operand *src_operand(int src)
{
    s_operand *operand;
    int		offset;

    operand = new_operand();
    if (src < 0)
    {
	operand->type = 2;
    	offset = src & 0x1fff;
        if (offset & 0x1000)
        {
	    offset = 0xfffff000 | offset;
        }
    	operand->immed = offset;
    }
    else
    {
	assert((src >= 0) && (src <= 63));
	operand->type = 1;
	operand->reg_num = src;
    }
    return (operand);
}

//**********
s_operand *address_operand(int src, char *p)
{
    s_operand *operand;
    int offset;
    int reg;
    char *str;
    int len;

    offset = src & 0x1fff;
    if (offset & 0x1000)
    {
	offset = 0xfffff000 | offset;
    }

    reg = (src >> 16) & 0x1f;
    assert((reg >= 0) && (reg <= 63));

    operand = new_operand();
    if (src < 0)
    {
    	len = strlen(p) + 3;
    	str = (char *)malloc(len * sizeof(char));
    	// SOLARIS: strcpy(str, "_");
    	strcpy(str, p);
	operand->type = 10;
	operand->reg_num = reg;
	operand->string = str;
    }
    else
    {
	operand->type = 9;
	operand->reg_num = reg;
	operand->immed = offset;
    }
    
    return (operand);
}

//**************************************
//**
//** Set target pointers of branches, jumps
//**
//**************************************

void label_pass()
{
    s_instr *siptr;
    label_linker *new_label;

    for (siptr = instr_head; siptr; siptr = siptr->next)
    {
	if (siptr->opcode == s_op_label)
	{
	    new_label = (label_linker *)malloc(sizeof(label_linker));
	    assert(new_label);
	    new_label->next = labels;
	    labels = new_label;
	    labels->label = siptr->src1->string;
	    labels->instr = siptr;
	}
    }

    return;
}

//**************
void branch_pass()
{
    s_instr 		*siptr;
    label_linker 	*lab;
    char 		*string;
    unsigned		len;

    for (siptr = instr_head; siptr; siptr = siptr->next)
    {
	switch (siptr->opcode)
	{
            // case s_op_jmp:
            case s_op_b:
            case s_op_be:
            case s_op_bne:
            case s_op_bl:
            case s_op_ble:
            case s_op_bleu:
            case s_op_bg:
            case s_op_bge:
            case s_op_bgu:
            case s_op_b_a:
            case s_op_bne_a:
            case s_op_bl_a:
            case s_op_ble_a:
            case s_op_bleu_a:
            case s_op_fbe:
            case s_op_fbne:
            case s_op_fbl:
            case s_op_fble:
            case s_op_fbg:
            case s_op_fbge:
            case s_op_fble_a:
            case s_op_fbne_a:
            case s_op_fbl_a:
		assert(siptr->src1);
		assert(siptr->src1->string);
		string = siptr->src1->string;
		for (lab = labels; lab; lab = lab->next)
		{
		    len = strlen(string);
		    if (len + 1 == strlen(lab->label))
		    if (!strncmp(lab->label, string, len))
			break;
		}
		assert(lab);
		siptr->target = lab->instr;
		break;
	    default:
		break;
	}
    }

}

