////////////////////////////////////////////////////////////////////////
//
//      Quinn Jacobson (qjacobso@ece.wisc.edu)
//      ECE department U of Wisconsin
//
//      CS 701
//      April 1996
//
//      instr_type.h
//
////////////////////////////////////////////////////////////////////////
//
// Defination of instr type
//
////////////////////////////////////////////////////////////////////////
 
//----- TYPEDEFs -----
typedef struct s_operand
{
    int 		type;	
	/*
		00: string	"%s", string
		01: reg 	"%%r%d", reg_num
		02: immed	"%d", immed
		03: lo str	"%%lo(_%s)", string
		04: lo imm  	"%%lo(%d)", immed
		05: lo com	"%%lo(_%s+%d)", string, immed
		06: hi str	"%%hi(_%s)", string
		07: hi imm  	"%%hi(%d)", immed
		08: hi com	"%%hi(_%s+%d)", string, immed
		09: addr	"[%%r%d+%d]", reg_num, immed
		10: addr lo str "[%%r%d+%%lo(%s)]", reg_num, string
		11: addr lo imm "[%%r%d+%%lo(%d)]", reg_num, immed
		12: addr lo com	"[%%r%d+%%lo(%s+%d)]", reg_num, string, immed
		13: reg+immed	"%%r%d+%d", reg_num, immed
	*/	
    int			reg_num;
	/*
		0-31: int registers
		32-63: fp registers 0-31
	*/
    int			immed;
    char		*string;
} s_operand;

typedef struct s_instr
{
    int			opcode;

    struct s_operand	*src1;
    struct s_operand	*src2;
    struct s_operand	*dst;

    struct s_instr	*next;
    struct s_instr	*prev;

    struct s_instr	*target;

} s_instr;


// 51 =
// bit 0-31:    integer regs
// bit 32-47:   fp double regs
// bit 48:      cond code
// bit 49:      memory
// bit 50:      y reg
#define HAZARD_VECTOR_SIZE	51


//----- DEFINEs ------
#define s_op_unimp	1
#define s_op_label	2
#define s_op_nop	3
#define s_op_ldsb	4
#define s_op_ldub	5
#define	s_op_ldsh	6
#define s_op_lduh	7
#define s_op_ld		8
#define s_op_ldd	9
#define s_op_stb	10
#define s_op_sth	11
#define s_op_st		12
#define s_op_std	13
#define s_op_cmp	14
#define s_op_sethi	15
#define s_op_mov	16
#define s_op_and	17
#define s_op_or		18
#define s_op_xor	19
#define s_op_xnor	20
#define s_op_sra	21
#define s_op_srl	22
#define s_op_sll	23
#define s_op_add	24
#define s_op_sub	25
#define s_op_udiv	26
#define s_op_sdiv	27
#define s_op_umul	28
#define s_op_smul	29
#define s_op_wr_y	30
#define s_op_call	31
#define s_op_call_rem	32
#define s_op_fcmpd	33
#define s_op_fcmps	34
#define s_op_fmovs	35
#define s_op_fnegs	36
#define s_op_fmuld	37
#define s_op_fmuls	38
#define s_op_fdivd	39
#define s_op_fdivs	40
#define s_op_faddd	41
#define s_op_fadds	42
#define s_op_fsubd	43
#define s_op_fsubs	44
#define s_op_fitod	45
#define s_op_fitos	46
#define s_op_fdtoi	47
#define s_op_fstoi	48
#define s_op_fstod	49
#define s_op_fdtos	50

#define s_op_b		51
#define s_op_b_a	52
#define s_op_be		53
#define s_op_be_a	54	// NEW
#define s_op_bne	55
#define s_op_bne_a	56
#define s_op_bl		57
#define s_op_bl_a	58
#define s_op_bg		59
#define s_op_bg_a	60	// NEW
#define s_op_bgu	61
#define s_op_bgu_a	62	// NEW
#define s_op_bge	63
#define s_op_bge_a	64	// NEW
#define s_op_ble	65
#define s_op_ble_a	66
#define s_op_bleu	67
#define s_op_bleu_a	68

#define s_op_fbe	69
#define s_op_fbe_a	70	// NEW
#define s_op_fbne	71
#define s_op_fbne_a	72
#define s_op_fbl	73
#define s_op_fbl_a	74
#define s_op_fbge	75
#define s_op_fbge_a	76	// NEW
#define s_op_fbg	77
#define s_op_fbg_a	78	// NEW
#define s_op_fble	79
#define s_op_fble_a	80

#define s_op_jmp	81
#define s_op_ret	82
#define s_op_restore	83

#define s_op_call_div	84
