////////////////////////////////////////////////////////////////////////
//
//      Quinn Jacobson (qjacobso@ece.wisc.edu)
//      ECE department U of Wisconsin
//
//      CS 701
//      Feb 1996
//
//      operand.h
//
////////////////////////////////////////////////////////////////////////
//
// Handles Operands
//
// UPDATED: to recognize machine registers as operands
//
////////////////////////////////////////////////////////////////////////

//----- Includes -----
 
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

extern "C" {
#include <simple.h>
}

#include "reg_alloc.h"
#include "LocalVars.h"
#include "temp.h"
#include "instr.h"
#include "reg_class.h"
#include "operands.h"

//----- Globals -----

extern reg_class inreg;
extern reg_class ccreg;
extern reg_class fpreg;

// UGLY HACK, no longer used
int mcpy_size;

//----- Code -----

//***************
REGnode *getreg(TEMPnode *temp, int is_load)
// get a integer type register for temp temp
{
    if (temp->temp >= big_number)
    {
	return (inreg.getreg(temp, is_load));
    }
    else
    {
	return (ccreg.getreg(temp, is_load));
    }
}

int reg_num(TEMPnode *temp, int def)
{
    if (temp->fp)
    {

        if (def == 2) def = 2;
        else if (def == 3) def = 4;
        else if (def == 24) def = 0;
        else def = 30;
    }
    
    if (temp->fp)
	return (fpreg.REGreg(temp->reg, def));
    else if (temp->temp >= big_number) 
	return (inreg.REGreg(temp->reg, def));
    else
	return (ccreg.REGreg(temp->reg, def));
}

void freereg(TEMPnode *temp)
{
    if (temp->fp)
	fpreg.freereg(temp->reg);
    else if (temp->temp >= big_number)
	inreg.freereg(temp->reg);
    else
	ccreg.freereg(temp->reg);
    return;
}

//***************
int getDST(int is_load, simple_reg *dst)
// Arrange for a destination operand
{
    TEMPnode *tptr;
    int		temp;
    int		is_fp;

    if (dst->kind == MACHINE_REG)
    {
	return (dst->num);	
    }
    if (dst->kind != TEMP_REG)
    {
	fprintf(stderr,"-- ERROR -- getDSTreg(): not a TEMP_REG!\n");
    }

    temp = dst->num;
    is_fp = (dst->var->type->base == FLOAT_TYPE);
 
    tptr = newTEMP(temp);
    tptr->type = 2;
    tptr->fp = is_fp;
    tptr->size = mcpy_size;
    if (!is_fp) 
    {
	tptr->reg = getreg(tptr, is_load);
        return (reg_num(tptr, 0));
    }
    else       
    {
	tptr->reg = fpreg.getreg(tptr, 0);
        return (reg_num(tptr, 0) | 0x20);
    }
}

//***************
int pseudoSRC(simple_reg *src, int reg)
// handle the case that a source operand is a pseudo register
{
    int	     offset;
    int      r_reg;
    int     c_reg;
    s_instr  *siptr;

    if (src->var->type->base == FLOAT_TYPE)
    {

        if (reg == 2) r_reg = 2;
        else if (reg == 3) r_reg = 4;
        else if (reg == 24) r_reg = 0;
        else r_reg = 30;
	c_reg = 32;
    }
    else
    {
       r_reg = reg;
       c_reg = 0;
    }
    offset = LVoffset(src->var->name);	
    //printf("!pseudoSRC() : %s\n",src->var->name);
    if (offset != 0)
    {
	    switch (src->var->type->len)
	    {
		case 8 :
		    if (src->var->type->base == SIGNED_TYPE)
		    {
		 	siptr = new_instr(s_op_ldsb);
			siptr->src1 = fp_offset_operand(offset);
			siptr->dst = reg_operand(r_reg);
		    }
		    else	
		    {
		 	siptr = new_instr(s_op_ldub);
			siptr->src1 = fp_offset_operand(offset);
			siptr->dst = reg_operand(r_reg);
		    }
		    break;
		case 16 :
		    if (src->var->type->base == SIGNED_TYPE)
		    {
		 	siptr = new_instr(s_op_ldsh);
			siptr->src1 = fp_offset_operand(offset);
			siptr->dst = reg_operand(r_reg);
		    }
		    else	
		    {
		 	siptr = new_instr(s_op_lduh);
			siptr->src1 = fp_offset_operand(offset);
			siptr->dst = reg_operand(r_reg);
		    }
		    break;
		case 32 :
		    siptr = new_instr(s_op_ld);
		    siptr->src1 = fp_offset_operand(offset);
		    siptr->dst = reg_operand(c_reg + r_reg);
		    break;
		case 64 :
		
		    if (offset % 8)
		    {
		        siptr = new_instr(s_op_ld);
		        siptr->src1 = fp_offset_operand(offset);
		        siptr->dst = reg_operand(c_reg + r_reg);
		        siptr = new_instr(s_op_ld);
		        siptr->src1 = fp_offset_operand(offset+4);
		        siptr->dst = reg_operand(c_reg + r_reg + 1);
		    }
		    else
		    {
		        siptr = new_instr(s_op_ldd);
		        siptr->src1 = fp_offset_operand(offset);
		        siptr->dst = reg_operand(c_reg + r_reg);
		    }
		    break;
		default :
		    assert(0);
	}
	if (c_reg == 32)
            return(0x20 | r_reg);
        else
	    return(r_reg);
    }
    fprintf(stderr,"-- ERROR -- pseudoSRC(): PsudoREG not a stack variable!\n");
    exit (1);
}

//***************
// ADDRESS ENCODING
// bit 31 : use symbol
// bit 30 : 8 byte alligned
// bit 20-16 : reg #
// bit 12-0 : offset
int addressSRC(simple_reg *src, char **sym)
// Arrange for an address operand
{
    int		reg;
    int		offset;
    TEMPnode	*tptr;
    int		is_8;
    s_instr  *siptr;

    is_8 = 0;
    if (src->kind == MACHINE_REG)
    {
	reg = src->num;
	offset = 0;
	is_8 = 0;
    }
    else if (src->kind == PSEUDO_REG)
    {
	reg = pseudoSRC(src, 4);
	offset = 0;
	is_8 = 0;
    }
    else if (src->kind == TEMP_REG)
    {
	tptr = findTEMP(src->num);
	switch (tptr->type)
	{
	    case 1 :
		offset = tptr->immed;
		reg = 0;
		if (offset % 8) is_8 = 0; else is_8 = 1;
		break;
	    case 2 :
		offset = 0;
		reg = reg_num(tptr, 4);
		is_8 = 0;
		break;
	    case 3 :
		offset = tptr->immed;
		reg = reg_num(tptr, 4);
		is_8 = 0;
		break;
	    case 4:
		offset = tptr->immed;
		reg = 30;
		if (offset % 8) is_8 = 0; else is_8 = 1;
		break;
	    case 5 :
		offset = 0xffff;
		*sym = tptr->sym;
		reg = reg_num(tptr, 4);
		is_8 = 1;
		break;
	    case 6 :
		offset = 0xffff;
		*sym = tptr->sym;
		siptr = new_instr(s_op_sethi);
		siptr->src1 = hi_operand(*sym, 0);
		siptr->dst = reg_operand(4);
		reg = 4;
		is_8 = 1;
		break;
	    default:
	       fprintf(stderr,"-- ERROR -- AddressSRC(): Illegal temp type!\n");
	       exit (1);
	}
	removeTEMP(tptr);
    }
    else
    {
	fprintf(stderr,"-- ERROR -- AddressSRC(): Illegal reg type!\n");
	exit (1);
    }
    if (offset == 0xffff)
    {
	assert (reg);
	return( (1<<31)|(reg<<16)|(is_8 << 30) );
    }
    else
    {
	assert (reg);
	return( (reg<<16)|(offset & 0x1fff)|(is_8 << 30) );
    }
}

//***************
// SRC return type:
// 	machine register number OR
//	0x10000immed
int RegSRC(simple_reg *src, int work)
// work is which work register to use : Op1 = 2, Op2 = 3
// Arrange for a source operand that has to be a register.
{
    int		reg;
    int		offset;
    TEMPnode	*tptr;
    char	*sym;
    s_instr  *siptr;

    if (src->kind == MACHINE_REG)
    {
	reg = src->num;
    }
    else if (src->kind == PSEUDO_REG)
    {
	reg = pseudoSRC(src, work);
    }
    else if (src->kind == TEMP_REG)
    {
	tptr = findTEMP(src->num);
	// UGLY HACK
   	if (tptr->size)
	    mcpy_size = tptr->size;
	if (tptr->fp)
	{
	    reg = reg_num(tptr, work) | 0x20;
	}
	else
	{
	    switch (tptr->type)
	    {
	        case 1 :
		    offset = tptr->immed;
		    siptr = new_instr(s_op_or);
		    siptr->src1 = zero_operand();
		    siptr->src2 = immed_operand(offset);
		    siptr->dst = reg_operand(work);
		    reg = work;
		    break;
	        case 2 :
		    reg = reg_num(tptr, work);
		    break;
	        case 3 :
		    offset = tptr->immed;
		    reg = reg_num(tptr, work);
		    siptr = new_instr(s_op_add);
		    siptr->src1 = reg_operand(reg);
		    siptr->src2 = immed_operand(offset);
		    siptr->dst = reg_operand(reg);
		    break;
	        case 4:
		    offset = tptr->immed;
		    reg = work;
		    siptr = new_instr(s_op_add);
		    siptr->src1 = reg_operand(30);
		    siptr->src2 = immed_operand(offset);
		    siptr->dst = reg_operand(reg);
		    break;
	        case 5 :
		    sym = tptr->sym;
		    reg = reg_num(tptr, work);
		    siptr = new_instr(s_op_or);
		    siptr->src1 = reg_operand(reg);
		    siptr->src2 = lo_operand(sym, 0);
		    siptr->dst = reg_operand(reg);
		    break;
	        case 6 :
		    sym = tptr->sym;
		    reg = work;
		    siptr = new_instr(s_op_sethi);
		    siptr->src1 = hi_operand(sym, 0);
		    siptr->dst = reg_operand(reg);
		    siptr = new_instr(s_op_or);
		    siptr->src1 = reg_operand(reg);
		    siptr->src2 = lo_operand(sym, 0);
		    siptr->dst = reg_operand(reg);
		    break;
	        case 7:
		    offset = tptr->immed;
		    reg = work;
		    siptr = new_instr(s_op_add);
		    siptr->src1 = reg_operand(14);
		    siptr->src2 = immed_operand(offset);
		    siptr->dst = reg_operand(reg);
		    break;
	        default:
	           fprintf(stderr,"-- ERROR -- RegSRC(): Illegal temp type!\n");
	           exit (1);
	    }
	}
	removeTEMP(tptr);
    }
    else
    {
	fprintf(stderr,"-- ERROR -- RegSRC(): Illegal reg type!\n");
	exit (1);
    }
    assert (reg);
    return (reg);
}

//***************
int RegImmedSRC(simple_reg *src, int work)
// work is which work register to use : Op1 = 2, Op2 = 3
// Arrange for a source operand that can be an operand or a small constant
// In the case of FLOATING POINT the operand must be in a register
{
    int		reg;
    int		offset;
    TEMPnode	*tptr;
    char	*sym;
    s_instr  *siptr;

    if (src->kind == MACHINE_REG)
    {
	reg = src->num;
    }
    else if (src->kind == PSEUDO_REG)
    {
	reg = pseudoSRC(src, work);
    }
    else if (src->kind == TEMP_REG)
    {
	tptr = findTEMP(src->num);
	// UGLY HACK
   	if (tptr->size)
	    mcpy_size = tptr->size;
	if (tptr->fp)
	{
	    reg = reg_num(tptr, work) | 0x20;
	}
	else
	{
	    switch (tptr->type)
	    {
	        case 1 :
		    offset = tptr->immed;
		    reg = (1<<31) | (offset & 0x1fff);
		    break;
	        case 2 :
		    reg = reg_num(tptr, work);
		    break;
	        case 3 :
		    offset = tptr->immed;
		    reg = reg_num(tptr, work);
		    siptr = new_instr(s_op_add);
		    siptr->src1 = reg_operand(reg);
		    siptr->src2 = immed_operand(offset);
		    siptr->dst = reg_operand(reg);
		    break;
	        case 4:
		    offset = tptr->immed;
		    reg = work;
		    siptr = new_instr(s_op_add);
		    siptr->src1 = reg_operand(30);
		    siptr->src2 = immed_operand(offset);
		    siptr->dst = reg_operand(reg);
		    break;
	        case 5 :
		    sym = tptr->sym;
		    reg = reg_num(tptr, work);
		    siptr = new_instr(s_op_or);
		    siptr->src1 = reg_operand(reg);
		    siptr->src2 = lo_operand(sym, 0);
		    siptr->dst = reg_operand(reg);
		    break;
	        case 6 :
		    sym = tptr->sym;
		    reg = work;
		    siptr = new_instr(s_op_sethi);
		    siptr->src1 = hi_operand(sym, 0);
		    siptr->dst = reg_operand(reg);
		    siptr = new_instr(s_op_or);
		    siptr->src1 = reg_operand(reg);
		    siptr->src2 = lo_operand(sym, 0);
		    siptr->dst = reg_operand(reg);
		    break;
	        default:
	           fprintf(stderr,"-- ERROR -- RegSRC(): Illegal temp type!\n");
	           exit (1);
	    }
	}
	removeTEMP(tptr);
    }
    else
    {
	fprintf(stderr,"-- ERROR -- RegSRC(): Illegal reg type!\n");
	exit (1);
    }
    assert (reg);
    return (reg);
}

//***************
int CallTarget(simple_reg *src, char **sym)
// Arrange for a call target
// If the target is a symbol, *sym will be set to the name and -1 returned
// Else a register value will be returned
{
    TEMPnode *tptr;

    if (src->kind == MACHINE_REG)
    {
	return (src->num);
    }
    else if (src->kind == PSEUDO_REG)
    {
	return (pseudoSRC(src, 2));
    }
    else if (src->kind == TEMP_REG)
    {
	tptr = findTEMP(src->num);
	if (tptr->type == 6)
	{
	    //printf("! tptr->sym = %d", tptr->sym); fflush(stdout);
	    *sym = tptr->sym;
	    return (-1);
	}
	else
	{
	    return (RegSRC(src, 2));
	}

    }
    else
    {
	fprintf(stderr,"-- ERROR -- CallTarget(): Illegal reg type!\n");
	exit (1);
    }
}

