////////////////////////////////////////////////////////////////////////
//
//      Quinn Jacobson (qjacobso@ece.wisc.edu)
//      ECE department U of Wisconsin
//
//      CS 701
//      Feb 1996
//
//      proc_instr.h
//
////////////////////////////////////////////////////////////////////////
//
// Traverses a functions instruction list and generates assembly code for
// the suif instructions.
//
// UPDATED: to optimize for set instruction followed by branch instruction
//
////////////////////////////////////////////////////////////////////////
 
 
//----- INCLUDE FILES -----

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

extern "C" {
#include <simple.h>
}

#include "ss_symtab.h"
#include "LocalVars.h"
#include "temp.h"
#include "operands.h"
#include "instr.h"
#include "reg_class.h"

//----- GLOBALS -----

extern reg_class fpreg;
extern reg_class ccreg;
extern reg_class inreg;

extern unsigned int ret_record_offset[256];
extern unsigned int ret_record_index;
extern unsigned int fp_save_offset;

// UGLY HACK no longer used
extern int mcpy_size;

// a easy way to generate unique labels
int quinn = 0;

// the sparc instruction list
extern s_instr *instr_head;
extern s_instr *instr_tail;

//----- CODE ----- 

//********************
void handle_base(simple_instr *instr, char *proc_name)
// translate BASE_FORM (most instructions) instructins
{
    int	src1, src2, dst;
    char name[80];
    char *p;
    int offset;
    TEMPnode *tptrNew;
    TEMPnode *tptrOld;
    int rec_size, trn_size;
    s_instr *siptr;
    char temp_str[80];

    name[0] = '\0';
    p = &(name[0]);

    switch(instr->opcode)
    {
	case NOP_OP :
	    break;
	case LOAD_OP :
	    if (instr->type->base == RECORD_TYPE)
	    {
		assert (instr->u.base.src1->kind == TEMP_REG);
		assert (instr->u.base.dst->kind == TEMP_REG);
		tptrNew=newTEMP(instr->u.base.dst->num);
		tptrOld = findTEMP(instr->u.base.src1->num);
		tptrNew->fp = 0;
		tptrNew->type = tptrOld->type;
		tptrNew->immed = tptrOld->immed;
		tptrNew->sym = tptrOld->sym;
		tptrNew->reg = tptrOld->reg;
		tptrOld->reg = NULL;
		removeTEMP(tptrOld);
		break;
	    }
	    src1 = addressSRC(instr->u.base.src1, &p);
	    dst = getDST(1,instr->u.base.dst);
	    switch(instr->type->len)
	    {
		case 8 :
		    if (instr->type->base == SIGNED_TYPE)
			siptr = new_instr(s_op_ldsb);
		    else
			siptr = new_instr(s_op_ldub);
	    	    siptr->src1 = address_operand( src1, p); 
	    	    siptr->dst = reg_operand( dst); 
		    break;
		case 16 :
		    if (instr->type->base == SIGNED_TYPE)
			siptr = new_instr(s_op_ldsh);
		    else
			siptr = new_instr(s_op_lduh);
	    	    siptr->src1 = address_operand( src1, p); 
	    	    siptr->dst = reg_operand( dst); 
		    break;
		case 32 :
		    siptr = new_instr(s_op_ld);
	    	    siptr->src1 = address_operand( src1, p); 
	    	    siptr->dst = reg_operand( dst); 
		    break;
		case 64 :
		    if (src1 & 0x40000000)
		    {
		        siptr = new_instr(s_op_ldd);
	    	        siptr->src1 = address_operand( src1, p); 
	    	        siptr->dst = reg_operand( dst); 
		    }
		    else
		    {
		        siptr = new_instr(s_op_ld);
	    	        siptr->src1 = address_operand( src1, p); 
	    	        siptr->dst = reg_operand( dst); 
		        siptr = new_instr(s_op_ld);
	    	        siptr->src1 = address_operand( src1+4, p); 
	    	        siptr->dst = reg_operand( dst+1); 
		    }
		    break;
		default :
		    assert(0);
	    }
	    break;
	case STR_OP :
	    if (instr->u.base.src2->var->type->base == RECORD_TYPE)
	    {
		assert (instr->u.base.src2->kind == TEMP_REG);
	        src1 = RegSRC(instr->u.base.src1, 2);
	        src2 = RegSRC(instr->u.base.src2, 3);
		rec_size = instr->u.base.src2->var->type->len >> 3;
		for (trn_size = 0; trn_size < rec_size; trn_size++)
		{
		    siptr = new_instr(s_op_ldub);
		    siptr->src1 = simple_address(src2, trn_size);
		    siptr->dst = reg_operand(1);
		    siptr = new_instr(s_op_stb);
		    siptr->src1 = reg_operand(1);
		    siptr->dst = simple_address(src1, trn_size);
		}
		break;
	    }
	    src2 = RegSRC(instr->u.base.src2, 3);
	    src1 = addressSRC(instr->u.base.src1, &p);
	    switch(instr->u.base.src2->var->type->len)
	    {
		case 8 :
	  	    siptr = new_instr(s_op_stb);
	    	    siptr->src1 = src_operand( src2); 
	    	    siptr->dst = address_operand( src1, p); 
		    break;
		case 16 :
	  	    siptr = new_instr(s_op_sth);
	    	    siptr->src1 = src_operand( src2); 
	    	    siptr->dst = address_operand( src1, p); 
		    break;
		case 32 :
		    siptr = new_instr(s_op_st);
	    	    siptr->src1 = src_operand( src2); 
	    	    siptr->dst = address_operand( src1, p); 
		    break;
		case 64 :
		    if (src1&0x40000000)
		    {
		        siptr = new_instr(s_op_std);
	    	        siptr->src1 = src_operand( src2); 
	    	        siptr->dst = address_operand( src1, p); 
		    }
		    else
		    {
		        siptr = new_instr(s_op_st);
	    	        siptr->src1 = src_operand( src2); 
	    	        siptr->dst = address_operand( src1, p); 
		        siptr = new_instr(s_op_st);
	    	        siptr->src1 = src_operand( src2+1); 
	    	        siptr->dst = address_operand( src1+4, p); 
		    }
		    break;
		default :
		    fprintf(stderr,"STR_OP : len = %d!\n", instr->type->len);
		    assert(0);
	    }
	    break;
	case NEG_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    dst = getDST(0,instr->u.base.dst);
	    if (instr->type->base == FLOAT_TYPE)
	    {
		siptr = new_instr(s_op_fnegs); siptr->src1 = src_operand( src1); 
		siptr->dst = reg_operand( dst); 
		if (instr->type->len == 64);
		{
		    siptr = new_instr(s_op_fmovs); siptr->src1 = src_operand( src1+1); 
		    siptr->dst = reg_operand( dst+1); 
		}
	    }
	    else
	    {
		siptr = new_instr(s_op_sub); 
		siptr->src1 = zero_operand();
		siptr->src2 = src_operand( src1); 
		siptr->dst = reg_operand( dst); 

	    }
	    break;
	case AND_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    dst = getDST(0,instr->u.base.dst);
	    assert (instr->type->base == UNSIGNED_TYPE);
	    siptr = new_instr(s_op_and); siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); siptr->dst = reg_operand( dst); 
	    break;
	case IOR_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    dst = getDST(0,instr->u.base.dst);
	    assert (instr->type->base == UNSIGNED_TYPE);
	    siptr = new_instr(s_op_or); siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); siptr->dst = reg_operand( dst); 
	    break;
	case XOR_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    dst = getDST(0,instr->u.base.dst);
	    assert (instr->type->base == UNSIGNED_TYPE);
	    siptr = new_instr(s_op_xor); siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); siptr->dst = reg_operand( dst); 
	    break;
	case ASR_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    dst = getDST(0,instr->u.base.dst);
	    assert (instr->type->base != FLOAT_TYPE);
	    siptr = new_instr(s_op_sra); siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); siptr->dst = reg_operand( dst); 
	    break;
	case LSR_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    dst = getDST(0,instr->u.base.dst);
	    assert (instr->type->base != FLOAT_TYPE);
	    siptr = new_instr(s_op_srl); siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); siptr->dst = reg_operand( dst); 
	    break;
	case LSL_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    dst = getDST(0,instr->u.base.dst);
	    assert (instr->type->base != FLOAT_TYPE);
	    siptr = new_instr(s_op_sll); siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); siptr->dst = reg_operand( dst); 
	    break;
	case NOT_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    dst = getDST(0,instr->u.base.dst);
	    assert (instr->type->base == UNSIGNED_TYPE);
	    siptr = new_instr(s_op_xnor); siptr->src1 = src_operand( src1); 
	    siptr->src2 = zero_operand();
	    siptr->dst = reg_operand( dst); 
	    break;
	case REM_OP :
	case MOD_OP :
	    src1 = RegSRC(instr->u.base.src1, 8);
	    src2 = RegSRC(instr->u.base.src2, 9);
	    dst = getDST(0,instr->u.base.dst);
	    assert (instr->type->base != FLOAT_TYPE);
	    if (src1 != 8) 
	    {
		siptr = new_instr(s_op_mov); siptr->src1 = src_operand( src1); 
		siptr->dst = reg_operand(8);
	    }
	    if (src2 != 9) 
	    {
		siptr = new_instr(s_op_mov); siptr->src1 = src_operand( src2); 
		siptr->dst = reg_operand(9);
	    }
	    siptr = new_instr(s_op_call_rem);
	    siptr->src1 = new_operand();
	    siptr->src1->type = 00;
	    siptr->src1->string = ".rem";
	    siptr->src2 = immed_operand(0);
	    siptr = new_instr(s_op_nop);
	    siptr = new_instr(s_op_mov); 
	    siptr->src1 = reg_operand(8);
	    siptr->dst = reg_operand( dst); 
	    break;
	case MUL_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    dst = getDST(0,instr->u.base.dst);
	    if (instr->type->base == FLOAT_TYPE)
	    {
		if(instr->type->len == 64)
	            siptr = new_instr(s_op_fmuld);
		else
		    siptr = new_instr(s_op_fmuls);
	    }
	    else
	    {
		if (instr->type->base == SIGNED_TYPE)
		    siptr = new_instr(s_op_smul);
		else
	            siptr = new_instr(s_op_umul);
	    }
	    siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); 
	    siptr->dst = reg_operand( dst); 
	    break;
	case DIV_OP :
	  if (instr->type->base == FLOAT_TYPE) {
	    // floating point division instruction works
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    dst = getDST(0,instr->u.base.dst);
	    if(instr->type->len == 64)
	      siptr = new_instr(s_op_fdivd);
	    else
	      siptr = new_instr(s_op_fdivs);
	    siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); 
	    siptr->dst = reg_operand( dst); 
	    break;
	  }
	  else {
	    // integer division instruction (sdiv) does NOT work if
	    // numerator is negative, so use call to .div instead
	    src1 = RegSRC(instr->u.base.src1, 8);
	    src2 = RegSRC(instr->u.base.src2, 9);
	    dst = getDST(0,instr->u.base.dst);
	    if (src1 != 8) {
	      siptr = new_instr(s_op_mov);
	      siptr->src1 = src_operand(src1);
	      siptr->dst = reg_operand(8);
	    }
	    if (src2 != 9) {
	      siptr = new_instr(s_op_mov);
	      siptr->src1 = src_operand( src2); 
	      siptr->dst = reg_operand(9);
	    }
	    siptr = new_instr(s_op_call_div);
	    siptr->src1 = new_operand();
	    siptr->src1->type = 00;
	    siptr->src1->string = ".div";
	    siptr->src2 = immed_operand(0);
	    siptr = new_instr(s_op_nop);
	    siptr = new_instr(s_op_mov); 
	    siptr->src1 = reg_operand(8);
	    siptr->dst = reg_operand( dst); 
	    break;
	  }
	case ADD_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    dst = getDST(0,instr->u.base.dst);
	    if (instr->type->base == FLOAT_TYPE)
	    {
		if(instr->type->len == 64)
	            siptr = new_instr(s_op_faddd);
		else		
	            siptr = new_instr(s_op_fadds);
	    }
	    else
	    {
	        siptr = new_instr(s_op_add);
	    }
	    siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); 
	    siptr->dst = reg_operand( dst); 
	    break;
	case SUB_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    dst = getDST(0,instr->u.base.dst);
	    if (instr->type->base == FLOAT_TYPE)
	    {
		if(instr->type->len == 64)
	            siptr = new_instr(s_op_fsubd);
		else
	            siptr = new_instr(s_op_fsubs);
	    }
	    else
	    {
	        siptr = new_instr(s_op_sub);
	    }
	    siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); 
	    siptr->dst = reg_operand( dst); 
	    break;
	case SEQ_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
	    {
		if(instr->u.base.src1->var->type->len == 64)
	            siptr = new_instr(s_op_fcmpd);
		else
	            siptr = new_instr(s_op_fcmps);
	    }
	    else
	        siptr = new_instr(s_op_cmp);
	    siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); 

	    // OPTIMIZE cmp/branch consecutive 
	    if (instr->next)
	    if ((instr->next->opcode == BTRUE_OP) || 
					(instr->next->opcode == BFALSE_OP))
	    if (instr->next->u.bj.src == instr->u.base.dst)
	    if (instr->next->u.bj.src->kind == TEMP_REG)
	    {
		if (instr->next->opcode == BTRUE_OP)
		{
	    	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
		    {
			new_instr(s_op_nop);
			siptr = new_instr(s_op_fbe);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
	    	    else
		    {
	        	siptr = new_instr(s_op_be);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
		}
		else
		{
	    	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
		    {
			new_instr(s_op_nop);
			siptr = new_instr(s_op_fbne);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
	    	    else
		    {
	        	siptr = new_instr(s_op_bne);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
		}
		new_instr(s_op_nop);
		instr->next = instr->next->next;
		break;
	    }

	    dst = getDST(0,instr->u.base.dst);
	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
	    {
		new_instr(s_op_nop);
		siptr = new_instr(s_op_fbne_a);
		sprintf(temp_str,"%s_%d", proc_name, quinn);
		siptr->src1 = jump_target("_quinn", temp_str);
	    }
	    else
	    {
	        siptr = new_instr(s_op_bne_a); 
		sprintf(temp_str,"%s_%d", proc_name, quinn);
		siptr->src1 = jump_target("_quinn", temp_str);
	    }
	    siptr = new_instr(s_op_add);
	    siptr->src1 = zero_operand();
	    siptr->src2 = zero_operand();
	    siptr->dst = reg_operand( dst); 
	    siptr = new_instr(s_op_add);
	    siptr->src1 = zero_operand();
	    siptr->src2 = immed_operand(1);
	    siptr->dst = reg_operand( dst); 
	    siptr = new_instr(s_op_label);
	    sprintf(temp_str,"%s_%d", proc_name, quinn);
	    siptr->src1 = quinn_label("_quinn", temp_str);
	    quinn++;
	    break;
	case SNE_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
	    {
		if(instr->u.base.src1->var->type->len == 64)
	            siptr = new_instr(s_op_fcmpd);
		else
	            siptr = new_instr(s_op_fcmps);
	    }
	    else
	        siptr = new_instr(s_op_cmp);
	    siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); 

	    // OPTIMIZE cmp/branch consecutive 
	    if (instr->next)
	    if ((instr->next->opcode == BTRUE_OP) || 
					(instr->next->opcode == BFALSE_OP))
	    if (instr->next->u.bj.src == instr->u.base.dst)
	    if (instr->next->u.bj.src->kind == TEMP_REG)
	    {
		if (instr->next->opcode == BTRUE_OP)
		{
	    	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
		    {
			new_instr(s_op_nop);
			siptr = new_instr(s_op_fbne);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
	    	    else
		    {
	        	siptr = new_instr(s_op_bne);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
		}
		else
		{
	    	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
		    {
			new_instr(s_op_nop);
			siptr = new_instr(s_op_fbe);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
	    	    else
		    {
	        	siptr = new_instr(s_op_be);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
		}
		new_instr(s_op_nop);
		instr->next = instr->next->next;
		break;
	    }

	    dst = getDST(0,instr->u.base.dst);
	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
	    {
		new_instr(s_op_nop);
		siptr = new_instr(s_op_fbne_a);
	        sprintf(temp_str,"%s_%d", proc_name, quinn);
		siptr->src1 = jump_target("_quinn", temp_str);
	    }
	    else
	    {
	        siptr = new_instr(s_op_bne_a);
	        sprintf(temp_str,"%s_%d", proc_name, quinn);
		siptr->src1 = jump_target("_quinn", temp_str);
	    }
	    siptr = new_instr(s_op_add);
	    siptr->src1 = zero_operand();
	    siptr->src2 = immed_operand(1);
	    siptr->dst = reg_operand( dst); 
	    siptr = new_instr(s_op_add);
	    siptr->src1 = zero_operand();
	    siptr->src2 = zero_operand();
	    siptr->dst = reg_operand( dst); 
	    siptr = new_instr(s_op_label);
	    sprintf(temp_str,"%s_%d", proc_name, quinn);
	    siptr->src1 = quinn_label("_quinn", temp_str);
	    quinn++;
	    break;
	case SL_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
	    {
		if(instr->u.base.src1->var->type->len == 64)
	            siptr = new_instr(s_op_fcmpd);
		else
	            siptr = new_instr(s_op_fcmps);
	    }
	    else
	        siptr = new_instr(s_op_cmp);
	    siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); 

	    // OPTIMIZE cmp/branch consecutive 
	    if (instr->next)
	    if ((instr->next->opcode == BTRUE_OP) || 
					(instr->next->opcode == BFALSE_OP))
	    if (instr->next->u.bj.src == instr->u.base.dst)
	    if (instr->next->u.bj.src->kind == TEMP_REG)
	    {
		if (instr->next->opcode == BTRUE_OP)
		{
	    	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
		    {
			new_instr(s_op_nop);
			siptr = new_instr(s_op_fbl);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
	    	    else
		    {
	        	siptr = new_instr(s_op_bl);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
		}
		else
		{
	    	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
		    {
			new_instr(s_op_nop);
			siptr = new_instr(s_op_fbge);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
	    	    else
		    {
	        	siptr = new_instr(s_op_bge);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
		}
		new_instr(s_op_nop);
		instr->next = instr->next->next;
		break;
	    }

	    dst = getDST(0,instr->u.base.dst);
	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
	    {
		new_instr(s_op_nop);
		siptr = new_instr(s_op_fbl_a);
	    	sprintf(temp_str,"%s_%d", proc_name, quinn);
		siptr->src1 = jump_target("_quinn", temp_str);
	    }
	    else
	    {
	        siptr = new_instr(s_op_bl_a); 
	    	sprintf(temp_str,"%s_%d", proc_name, quinn);
		siptr->src1 = jump_target("_quinn", temp_str);
	    }
	    siptr = new_instr(s_op_add);
	    siptr->src1 = zero_operand();
	    siptr->src2 = immed_operand(1);
	    siptr->dst = reg_operand( dst); 
	    siptr = new_instr(s_op_add);
	    siptr->src1 = zero_operand();
	    siptr->src2 = zero_operand();
	    siptr->dst = reg_operand( dst); 
	    siptr = new_instr(s_op_label);
	    sprintf(temp_str,"%s_%d", proc_name, quinn);
	    siptr->src1 = quinn_label("_quinn", temp_str);
	    quinn++;
	    break;
	case SLE_OP :
	    src1 = RegSRC(instr->u.base.src1, 2);
	    src2 = RegImmedSRC(instr->u.base.src2, 3);
	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
	    {
		if(instr->u.base.src1->var->type->len == 64)
	            siptr = new_instr(s_op_fcmpd);
		else
	            siptr = new_instr(s_op_fcmps);
	    }
	    else
	        siptr = new_instr(s_op_cmp);
	    siptr->src1 = src_operand( src1); 
	    siptr->src2 = src_operand( src2); 

	    // OPTIMIZE cmp/branch consecutive 
	    if (instr->next)
	    if ((instr->next->opcode == BTRUE_OP) || 
					(instr->next->opcode == BFALSE_OP))
	    if (instr->next->u.bj.src == instr->u.base.dst)
	    if (instr->next->u.bj.src->kind == TEMP_REG)
	    {
		if (instr->next->opcode == BTRUE_OP)
		{
	    	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
		    {
			new_instr(s_op_nop);
			siptr = new_instr(s_op_fble);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
	    	    else if (instr->u.base.src1->var->type->base == SIGNED_TYPE)
		    {
	        	siptr = new_instr(s_op_ble);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
		    else
		    {
	        	siptr = new_instr(s_op_bleu);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
		}
		else
		{
	    	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
		    {
			new_instr(s_op_nop);
			siptr = new_instr(s_op_fbg);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
	    	    else if (instr->u.base.src1->var->type->base == SIGNED_TYPE)
		    {
	        	siptr = new_instr(s_op_bg);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
		    else
		    {
	        	siptr = new_instr(s_op_bgu);
			siptr->src1 = jump_target(
				instr->next->u.bj.target->name,proc_name);
		    }
		}
		new_instr(s_op_nop);
		instr->next = instr->next->next;
		break;
	    }

	    dst = getDST(0,instr->u.base.dst);
	    if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
	    {
		new_instr(s_op_nop);
		siptr = new_instr(s_op_fble_a);
	    	sprintf(temp_str,"%s_%d", proc_name, quinn);
		siptr->src1 = jump_target("_quinn", temp_str);
	    }
	    else if (instr->u.base.src1->var->type->base == SIGNED_TYPE)
	    {
	        siptr = new_instr(s_op_ble_a); 
	    	sprintf(temp_str,"%s_%d", proc_name, quinn);
		siptr->src1 = jump_target("_quinn", temp_str);
	    }
	    else
	    {
	        siptr = new_instr(s_op_bleu_a); 
	    	sprintf(temp_str,"%s_%d", proc_name, quinn);
		siptr->src1 = jump_target("_quinn", temp_str);
	    }
	    siptr = new_instr(s_op_add);
	    siptr->src1 = zero_operand();
	    siptr->src2 = immed_operand(1);
	    siptr->dst = reg_operand( dst); 
	    siptr = new_instr(s_op_add);
	    siptr->src1 = zero_operand();
	    siptr->src2 = zero_operand();
	    siptr->dst = reg_operand( dst); 
	    siptr = new_instr(s_op_label);
	    sprintf(temp_str,"%s_%d", proc_name, quinn);
	    siptr->src1 = quinn_label("_quinn", temp_str);
	    quinn++;
	    break;
	case CPY_OP :
	    if (instr->u.base.dst->kind == MACHINE_REG)
	    {
		src1 = RegSRC(instr->u.base.src1, 2);
		siptr = new_instr(s_op_mov);
		siptr->src1 = src_operand( src1);
		siptr->dst = reg_operand(instr->u.base.dst->num);
	    }
	    else if (instr->u.base.dst->kind == PSEUDO_REG)
	    {
		src1 = RegSRC(instr->u.base.src1, 2);
		offset = LVoffset(instr->u.base.dst->var->name);
		assert(offset);
		if (instr->type->base == FLOAT_TYPE)
		{
		    if ((offset % 8) || (instr->type->len == 32))
		    {
		        siptr = new_instr(s_op_st);
		        siptr->src1 = src_operand( src1); 
			siptr->dst = fp_offset_operand(offset);

			if (instr->type->len == 64)
			{
		            siptr = new_instr(s_op_st);
		            siptr->src1 = src_operand( src1+1); 
			    siptr->dst = fp_offset_operand(offset + 4);
			}
		    }
		    else
		    {
		        siptr = new_instr(s_op_std);
		        siptr->src1 = src_operand( src1); 
			siptr->dst = fp_offset_operand(offset);
		    }
		}
		else
		{
		    switch(instr->u.base.dst->var->type->len)
		    {
			case 8 :
		    	    siptr = new_instr(s_op_stb);
			    break;
			case 16 :
 			    siptr = new_instr(s_op_sth);
			    break;
			case 32 :
		    	    siptr = new_instr(s_op_st);
			    break;
		    }
		    siptr->src1 = src_operand( src1); 
		    siptr->dst = fp_offset_operand(offset);
		}
	    }
	    else if (instr->u.base.dst->kind == TEMP_REG)
	    {
		src1 = RegSRC(instr->u.base.src1,2);
		dst = getDST(0,instr->u.base.dst);
		if (src1 != dst)	
		{
		    if (instr->type->base == FLOAT_TYPE)
		    {
		      // BUG FIX: The following line previously had "64"
		      //          instead of "32"
			if (instr->type->len == 32)
			{
			  // BUG FIX: The following line previously had
			  //          "s_op_mov" instead of "s_op_fmovs"
			    siptr = new_instr(s_op_fmovs); 
			    siptr->src1 = src_operand(src1);
			    siptr->dst = reg_operand( dst); 
			}
			else
			{
			    siptr = new_instr(s_op_fmovs); 
			    siptr->src1 = src_operand( src1); 
			    siptr->dst = reg_operand( dst); 

			    siptr = new_instr(s_op_fmovs); 
			    siptr->src1 = src_operand( src1+1); 
			    siptr->dst = reg_operand( dst+1); 
			}
		    }
		    else
		    {
			siptr = new_instr(s_op_mov); 
			siptr->src1 = src_operand( src1); 
			siptr->dst = reg_operand( dst); 
		    }
		}
	    }
	    else assert (0);
	    break;
	case RET_OP :
	    if (instr->u.base.src1)
	    {
	    if (instr->u.base.src1->var->type->base == RECORD_TYPE)
	    {
		siptr = new_instr(s_op_ld);
		siptr->src1 = fp_offset_operand(64);
		siptr->dst = reg_operand(24);
	 	if (instr->u.base.src1)
		{
		    src1 = RegSRC(instr->u.base.src1, 2);	
		    rec_size = instr->u.base.src1->var->type->len >> 3;
		    for (trn_size = 0; trn_size < rec_size; trn_size++)
		    {
			siptr = new_instr(s_op_ldub);
			siptr->src1 = simple_address(src1, trn_size);
			siptr->dst = reg_operand(1);
			siptr = new_instr(s_op_stb);
			siptr->src1 = reg_operand(1);
			siptr->dst = simple_address(24, trn_size);
		    }
		}
		siptr = new_instr(s_op_jmp);
		siptr->src1 = new_operand();
		siptr->src1->type = 13;
		siptr->src1->reg_num = 31;
		siptr->src1->immed = 12;
	    }
	    else
	    {
		src1 = RegSRC(instr->u.base.src1, 24);
		if (instr->u.base.src1->var->type->base == FLOAT_TYPE)
		{
		    if (src1 != 0x100)
		    {
			siptr = new_instr(s_op_fmovs); 
			siptr->src1 = src_operand( src1); 
			siptr->dst = reg_operand(32);
			if (instr->u.base.src1->var->type->len == 64)
			{
			    siptr = new_instr(s_op_fmovs); 
			    siptr->src1 = src_operand( src1+1); 
			    siptr->dst = reg_operand(33);
			}
		    }
		}
		else
		{
		    if (src1 != 24)
		    {
		        siptr = new_instr(s_op_mov);
		        siptr->src1 = src_operand( src1);
			siptr->dst = reg_operand(24);
		    }
		}
		siptr = new_instr(s_op_ret);
	    }
	    }
	    else
		siptr = new_instr(s_op_ret);
	    siptr = new_instr(s_op_restore);
	    break;
	case CVT_OP :
	    //assert(instr->u.base.dst->kind == TEMP_REG);
	    src1 = RegSRC(instr->u.base.src1,2);
	    dst = getDST(0,instr->u.base.dst);
	    if (((instr->u.base.src1->var->type->base == ADDRESS_TYPE) &&
	 	 (instr->u.base.dst->var->type->base == SIGNED_TYPE)    )  ||
		((instr->u.base.src1->var->type->base == ADDRESS_TYPE) &&
	 	 (instr->u.base.dst->var->type->base == UNSIGNED_TYPE)    ) ||
		((instr->u.base.src1->var->type->base == SIGNED_TYPE) &&
	 	 (instr->u.base.dst->var->type->base == ADDRESS_TYPE)    ) ||
		((instr->u.base.src1->var->type->base == ADDRESS_TYPE) &&
	 	 (instr->u.base.dst->var->type->base == ADDRESS_TYPE)    ) ||
		((instr->u.base.src1->var->type->base == UNSIGNED_TYPE) &&
	 	 (instr->u.base.dst->var->type->base == ADDRESS_TYPE)    )    )
	    {
		if (dst != src1)
		{
		    siptr = new_instr(s_op_mov);
		    siptr->src1 = src_operand( src1); 
		    siptr->dst = reg_operand( dst); 
		}
	    }
	    else
	    if	((instr->u.base.src1->var->type->base == SIGNED_TYPE) &&
	 	 (instr->u.base.dst->var->type->base == UNSIGNED_TYPE) )
	    {
		switch(instr->u.base.dst->var->type->len)
		{
		    case 8 :
			siptr = new_instr(s_op_sll); 
			siptr->src1 = src_operand( src1); 
			siptr->src2 = immed_operand(24);
			siptr->dst = reg_operand( dst); 
			siptr = new_instr(s_op_srl); 
			siptr->src1 = src_operand( dst); 
			siptr->src2 = immed_operand(24);
			siptr->dst = reg_operand( dst); 
			break;
		    case 16 :
			siptr = new_instr(s_op_sll); 
			siptr->src1 = src_operand( src1); 
			siptr->src2 = immed_operand(16);
			siptr->dst = reg_operand( dst); 
			siptr = new_instr(s_op_srl); 
			siptr->src1 = src_operand( dst); 
			siptr->src2 = immed_operand(16);
			siptr->dst = reg_operand( dst); 
			break;
		    case 32 :
			if (dst != src1)
			{
		    	    siptr = new_instr(s_op_mov);
		    	    siptr->src1 = src_operand( src1); 
		    	    siptr->dst = reg_operand( dst); 
			}
			break;
		}
	    }
	    else
	    if	((instr->u.base.src1->var->type->base == UNSIGNED_TYPE) &&
	 	 (instr->u.base.dst->var->type->base == SIGNED_TYPE)    )
	    {
		switch(instr->u.base.dst->var->type->len)
		{
		    case 8 :
			siptr = new_instr(s_op_sll); 
			siptr->src1 = src_operand( src1); 
			siptr->src2 = immed_operand(24);
			siptr->dst = reg_operand( dst); 
			siptr = new_instr(s_op_sra); 
			siptr->src1 = src_operand( dst); 
			siptr->src2 = immed_operand(24);
			siptr->dst = reg_operand( dst); 
			break;
		    case 16 :
			siptr = new_instr(s_op_sll); 
			siptr->src1 = src_operand( src1); 
			siptr->src2 = immed_operand(16);
			siptr->dst = reg_operand( dst); 
			siptr = new_instr(s_op_sra); 
			siptr->src1 = src_operand( dst); 
			siptr->src2 = immed_operand(16);
			siptr->dst = reg_operand( dst); 
			break;
		    case 32 :
			if (dst != src1)
			{
		    	    siptr = new_instr(s_op_mov);
		    	    siptr->src1 = src_operand( src1); 
		    	    siptr->dst = reg_operand( dst); 
			}
			break;
		}
	    }
	    else
	    if (((instr->u.base.src1->var->type->base == UNSIGNED_TYPE) &&
	 	 (instr->u.base.dst->var->type->base == UNSIGNED_TYPE)    ) ||
	    	((instr->u.base.src1->var->type->base == SIGNED_TYPE) &&
	 	 (instr->u.base.dst->var->type->base == SIGNED_TYPE)    ) )
	    {
		if (dst != src1)
		{
		    siptr = new_instr(s_op_mov);
		    siptr->src1 = src_operand( src1); 
		    siptr->dst = reg_operand( dst); 
		}
	    }
	    else
	    if (((instr->u.base.src1->var->type->base == UNSIGNED_TYPE) || 
	         (instr->u.base.src1->var->type->base == SIGNED_TYPE) ) &&
	 	 (instr->u.base.dst->var->type->base == FLOAT_TYPE)    )
	    {
		siptr = new_instr(s_op_st); 
		siptr->src1 = src_operand( src1); 
		siptr->dst = fp_offset_operand(-8);
		siptr = new_instr(s_op_ld);
		siptr->src1 = fp_offset_operand(-8);
		siptr->dst = reg_operand(62);
		if (instr->u.base.dst->var->type->len == 64)
		{
		    siptr = new_instr(s_op_fitod);
		    siptr->src1 = reg_operand(62);
		    siptr->dst = reg_operand( dst); 
		}
		else
		{
		    siptr = new_instr(s_op_fitos);
		    siptr->src1 = reg_operand(62);
		    siptr->dst = reg_operand( dst); 
		}
	    }
	    else
	    if ((instr->u.base.src1->var->type->base == FLOAT_TYPE) &&
	 	((instr->u.base.dst->var->type->base == SIGNED_TYPE) ||
	 	 (instr->u.base.dst->var->type->base == UNSIGNED_TYPE)))
	    {
		if (instr->u.base.src1->var->type->len == 64)
		{
		    siptr = new_instr(s_op_fdtoi); 
		    siptr->src1 = src_operand( src1); 
		    siptr->dst = reg_operand(62);
		}
		else
		{
		    siptr = new_instr(s_op_fstoi); 
		    siptr->src1 = src_operand( src1); 
		    siptr->dst = reg_operand(62);
		}
		siptr = new_instr(s_op_st);
		siptr->src1 = reg_operand(62);
		siptr->dst = fp_offset_operand(-8);
		siptr = new_instr(s_op_ld);
		siptr->src1 = fp_offset_operand(-8);
		siptr->dst = reg_operand( dst); 
	    }
	    else
	    if ((instr->u.base.src1->var->type->base == FLOAT_TYPE) &&
		(instr->u.base.dst->var->type->base == FLOAT_TYPE) )
	    {
		if (instr->u.base.src1->var->type->len == 32)
		{
		    assert(instr->u.base.dst->var->type->len == 64);
		    siptr = new_instr(s_op_fstod); 
		    siptr->src1 = src_operand( src1); 
		    siptr->dst = reg_operand( dst); 
		}
		else
		{
		    assert(instr->u.base.dst->var->type->len == 32);
		    siptr = new_instr(s_op_fdtos); 
		    siptr->src1 = src_operand( src1); 
		    siptr->dst = reg_operand( dst); 

		}
	    }
	    else assert (0);
	    break;
	case MCPY_OP :
	    src1 = RegSRC(instr->u.base.src1,2);
	    src2 = RegSRC(instr->u.base.src2,3);
	    /*
	    if (mcpy_size)
	        rec_size = mcpy_size;
	    else
		assert (0);
	    */
	    rec_size = get_referrent_size(instr->u.base.src2) >> 3;
	    if (rec_size % 4)
	    {
	        for(trn_size = 0; trn_size < rec_size; trn_size++)
	        {
		    siptr = new_instr(s_op_ldub);
		    siptr->src1 = simple_address(src2, trn_size);
		    siptr->dst = reg_operand(1);
		    siptr = new_instr(s_op_stb);
		    siptr->src1 = reg_operand(1);
		    siptr->dst = simple_address(src1, trn_size);
	        }
	    }
	    else
	    {
	        for(trn_size = 0; trn_size < rec_size; trn_size += 4)
	        {
		    siptr = new_instr(s_op_ld);
		    siptr->src1 = simple_address(src2, trn_size);
		    siptr->dst = reg_operand(1);
		    siptr = new_instr(s_op_st);
		    siptr->src1 = reg_operand(1);
		    siptr->dst = simple_address(src1, trn_size);
	        }
	    }
	    break;
	default:
	    assert (0);
	    break;
    }
    return;
}

//********************
void handle_bj(simple_instr *instr,char *proc_name)
// translate BJ_FORM (branch and jump) instructins
{
    int src;
    s_instr *siptr;

    switch (instr->opcode)
    {
	case JMP_OP :
	    siptr = new_instr(s_op_b);
	    siptr->src1 = jump_target(instr->u.bj.target->name,proc_name);
	    new_instr(s_op_nop);
	    break;
	case BTRUE_OP :
	    src = RegSRC(instr->u.bj.src, 2);
	    siptr = new_instr(s_op_cmp);
	    siptr->src1 = src_operand( src); siptr->src2 = zero_operand();
	    siptr = new_instr(s_op_bne);
	    siptr->src1 = jump_target(instr->u.bj.target->name,proc_name);
	    new_instr(s_op_nop);
	    break;
	case BFALSE_OP :
	    src = RegSRC(instr->u.bj.src, 2);
	    siptr = new_instr(s_op_cmp);
	    siptr->src1 = src_operand( src); siptr->src2 = zero_operand();
	    siptr = new_instr(s_op_be);
	    siptr->src1 = jump_target(instr->u.bj.target->name,proc_name);
	    new_instr(s_op_nop);
	    break;

	default : 
	    break;
    }
    return;
}

//********************
void handle_ldc(simple_instr *instr)
// translate LDC_FORM (load constant) instructins
{
    int		value, offset;
    TEMPnode*	tptr;
    int		temp;
    REGnode	*reg;
    char	*sym_name;
    int		rnum;
    s_instr	*siptr;

    //printf("handle_ldc()\n");
    if (instr->u.ldc.dst->kind != TEMP_REG)
    {
	fprintf(stderr,"-- ERROR -- handle_ldc(): dst is not a temp!\n");
	exit (1);
    }
    
    temp = instr->u.ldc.dst->num;
    tptr = newTEMP(temp);
    tptr->fp = 0;
    if (instr->u.ldc.value.format == IMMED_INT)
    {
	value = instr->u.ldc.value.u.ival;
	if ((value > 0xfff) || (value < -0xfff))
	{
	    reg = getreg(tptr,0);
	    siptr = new_instr(s_op_sethi);
	    siptr->src1 = hi_operand(NULL, value); 
	    siptr->dst = reg_operand(reg->num);
	    siptr = new_instr(s_op_or);
	    siptr->src1 = reg_operand(reg->num);
	    siptr->src2 = lo_operand(NULL, value);
	    siptr->dst = reg_operand(reg->num);
	    tptr->type = 2;
	    tptr->reg = reg;
	}
	else
	{
	    tptr->type = 1;
	    tptr->immed = value;
	}
	return;
    }
    if (instr->u.ldc.value.format == IMMED_SYMBOL)
    {
	value = instr->u.ldc.value.u.s.offset >> 3;
	sym_name = instr->u.ldc.value.u.s.symbol->name;
	offset = LVoffset(sym_name);
	if (offset == 0)
	{
	    if (value == 0)
	    {
	        tptr->type = 6;
	        tptr->sym = sym_name;
	    }
	    else
	    {
	        reg = getreg(tptr,0);
	        tptr->reg = reg;
		siptr = new_instr(s_op_sethi);
		siptr->src1 = hi_operand(sym_name, 0);
		siptr->dst = reg_operand(reg->num);
		siptr = new_instr(s_op_or);
		siptr->src1 = reg_operand(reg->num);
		siptr->src2 = lo_operand(sym_name, 0);
		siptr->dst = reg_operand(reg->num);
	 	if ((value < -0xfff) || (value > 0xfff))
		{
		    siptr = new_instr(s_op_sethi);
		    siptr->src1 = hi_operand(NULL, value);
		    siptr->dst = reg_operand(1);
		    siptr = new_instr(s_op_or);
		    siptr->src1 = reg_operand(1);
		    siptr->src2 = lo_operand(NULL, value);
		    siptr->dst = reg_operand(1);
		    siptr = new_instr(s_op_add);
		    siptr->src1 = reg_operand(1);
		    siptr->src2 = reg_operand(reg->num);
		    siptr->dst = reg_operand(reg->num);
		    tptr->type = 2;
		}
		else
		{
		    tptr->immed = value;
		    tptr->type = 3;
		}
	    }
	}
	else
	{
	    tptr->size = LVdata_size(sym_name);
	    value = value + offset; 
	    if ((value < -0xfff) || (value > 0xfff))
	    {
	        reg = getreg(tptr,0);
	        tptr->reg = reg;
		siptr = new_instr(s_op_add);
		siptr->src1 = reg_operand(30);
		siptr->src2 = lo_operand(NULL, value);
		siptr->dst = reg_operand(reg->num);
		siptr = new_instr(s_op_sethi);
		siptr->src1 = hi_operand(NULL, value);
		siptr->dst = reg_operand(1);
		siptr = new_instr(s_op_add);
		siptr->src1 = reg_operand(reg->num);
		siptr->src2 = reg_operand(1);
		siptr->dst = reg_operand(reg->num);

		tptr->type = 2;
	    }
	    else
	    {
		tptr->immed = value;
		tptr->type = 4;
	    }
	}
	return;
    }
    if (instr->u.ldc.value.format == IMMED_FLOAT)
    {
	double valued;
	float  valuef;
	int    lval, hval;
	int    *ptr;

        tptr->fp = 1;
	tptr->reg = fpreg.getreg(tptr, 0);
	if (instr->type->len == 64)
	{
	    valued = instr->u.ldc.value.u.fval;
	    ptr = (int*)(&valued);
	    hval = *ptr;
	    // BUG FIX replace 1 line (ptr = ptr++ with ptr++)
	    ptr++;
	    lval = *ptr;
	    siptr = new_instr(s_op_sethi);
	    siptr->src1 = hi_operand(NULL, hval);
	    siptr->dst = reg_operand(2);
	    siptr = new_instr(s_op_or);
	    siptr->src1 = reg_operand(2);
	    siptr->src2 = lo_operand(NULL, hval);
	    siptr->dst = reg_operand(2);

	    siptr = new_instr(s_op_sethi);
	    siptr->src1 = hi_operand(NULL, lval);
	    siptr->dst = reg_operand(3);
	    siptr = new_instr(s_op_or);
	    siptr->src1 = reg_operand(3);
	    siptr->src2 = lo_operand(NULL, lval);
	    siptr->dst = reg_operand(3);

	    siptr = new_instr(s_op_std);
	    siptr->src1 = reg_operand(2);
	    siptr->dst = fp_offset_operand(-8);
	    rnum = fpreg.REGreg(tptr->reg, 0);
	    siptr = new_instr(s_op_ldd);
	    siptr->src1 = fp_offset_operand(-8);
	    siptr->dst = reg_operand(rnum+32);
	}
	else
	{
	    valuef = instr->u.ldc.value.u.fval;
	    ptr = (int*)(&valuef);
	    hval = *ptr;
	    siptr = new_instr(s_op_sethi);
	    siptr->src1 = hi_operand(NULL, hval);
	    siptr->dst = reg_operand(2);
	    siptr = new_instr(s_op_or);
	    siptr->src1 = reg_operand(2);
	    siptr->src2 = lo_operand(NULL, hval);
	    siptr->dst = reg_operand(2);
	    siptr = new_instr(s_op_st);
	    siptr->src1 = reg_operand(2);
	    siptr->dst = fp_offset_operand(-8);
	    rnum = fpreg.REGreg(tptr->reg, 0);
	    siptr = new_instr(s_op_ld);
	    siptr->src1 = fp_offset_operand(-8);
	    siptr->dst = reg_operand(rnum+32);
	}
	return;
    }

    fprintf(stderr,"-- ERROR -- handle_ldc(): Unimplemented immediate type!\n");

}

//********************
void handle_call(simple_instr *instr)
// translate CALL_FORM (calls) instructins
{
    int i, r;
    int src;
    int dst;
    int nargs;
    simple_reg *arg;
    int target;
    char name[80];
    char *sym;
    REGnode *fpr;
    int rec_size, trn_size;
    TEMPnode *tptr;
    s_instr *siptr;

// HANDLE OUTGOINT PARAMS
    sym = &(name[0]);
    r = 8;
    nargs = instr->u.call.nargs;
    for (i = 0; i < nargs; i++)
    {
        arg = instr->u.call.args[i];
	switch (arg->var->type->base)
	{
	    case VOID_TYPE :
		assert (0);
		break;
	    case SIGNED_TYPE :
	    case UNSIGNED_TYPE :
	    case ADDRESS_TYPE :
	        if (r <= 13)
	        {
	            src = RegSRC(arg, r);
	            if (src != r)
	            {
		        siptr = new_instr(s_op_mov);
		        siptr->src1 = src_operand( src);
			siptr->dst = reg_operand(r);
	            }
	            r++;
	        }
	        else
	        {
	            src = RegSRC(arg, 2);
		    // BUG FIX: Use appropriate store instruction
		    //          (do not always used "st").
		    //          Single line replaced with switch. 
		     switch(arg->var->type->len){
                      case 8: // byte
                        siptr = new_instr(s_op_stb);
                        break;
                      case 16: // halfword
                        siptr = new_instr(s_op_sth);
                        break;
                      default: // word (should be 32)
                        siptr = new_instr(s_op_st);
                        break;
                    }

	            siptr->src1 = src_operand( src);
		    siptr->dst = sp_offset_operand(68+((r-8)*4));
		    r++;
	        }
		break;
	    case FLOAT_TYPE :
		src = RegSRC(arg, 2);
		if (arg->var->type->len == 64)
		{ // 64bit float
		    if (r <= 12)
		    {
		        siptr = new_instr(s_op_std);
		        siptr->src1 = src_operand( src);
			siptr->dst = fp_offset_operand(-8);
		        if (r % 2)
		        {
			    siptr = new_instr(s_op_ld);
			    siptr->src1 = fp_offset_operand(-8);
			    siptr->dst = reg_operand(r);
			    r++;
			    siptr = new_instr(s_op_ld);
			    siptr->src1 = fp_offset_operand(-4);
			    siptr->dst = reg_operand(r);
			    r++;
		        }
		        else
		        {
			    siptr = new_instr(s_op_ldd);
			    siptr->src1 = fp_offset_operand(-8);
			    siptr->dst = reg_operand(r);
			    r += 2;
		        }
		    }
		    else if (r == 13)
		    {
		        siptr = new_instr(s_op_std);
		        siptr->src1 = src_operand( src);
			siptr->dst = sp_offset_operand(88);
			siptr = new_instr(s_op_ld);
			siptr->src1 = sp_offset_operand(88);
			siptr->dst = reg_operand(13);
		        r += 2;
		    }
		    else
		    {
		        if ((68+(4*(r-8))) % 8)
		        {
		            siptr = new_instr(s_op_st);
		            siptr->src1 = src_operand( src);
			    siptr->dst = sp_offset_operand(68+(4*(r-8)));
			    r++;
		            siptr = new_instr(s_op_st);
		            siptr->src1 = src_operand( src+1);
			    siptr->dst = sp_offset_operand(68+(4*(r-8)));
			    r++;
		        }
		        else
		        {
		            siptr = new_instr(s_op_std);
		            siptr->src1 = src_operand( src);
			    siptr->dst = sp_offset_operand(68+(4*(r-8)));
		            r += 2;
		        }
		    }
		}
		else
		{ // 32bit float
		    if (r <= 13)
		    {
		        siptr = new_instr(s_op_st);
		        siptr->src1 = src_operand( src);
			siptr->dst = fp_offset_operand(-8);
			siptr = new_instr(s_op_ld);
			siptr->src1 = fp_offset_operand(-8);
			siptr->dst = reg_operand(r);
			r++;
		    }			
		    else
		    {
		        siptr = new_instr(s_op_st);
		        siptr->src1 = src_operand( src);
			siptr->dst = sp_offset_operand(68+(4*(r-8)));
			r++;
		    }

		}
		break;
	    case RECORD_TYPE :
		src = RegSRC(arg, 2);
		rec_size = arg->var->type->len >> 3;
		for (trn_size = 0; trn_size < rec_size; trn_size += 4)
		{
		    if (r <= 13)
		    {
			siptr = new_instr(s_op_ld);
			siptr->src1 = simple_address(src, trn_size);
			siptr->dst = reg_operand(r);
		    }
		    else
		    {
			siptr = new_instr(s_op_ld);
			siptr->src1 = simple_address(src, trn_size);
			siptr->dst = reg_operand(1);
			siptr = new_instr(s_op_st); 
			siptr->src1 = reg_operand(1);
			siptr->dst = sp_offset_operand(68+(4*(r-8)));
		    }
		    r++;
		}
		break;
	}
	fflush(stdout);
    }
// Save fp registers
    assert (inreg.reglist() == NULL);
    for (fpr = fpreg.reglist(); fpr; fpr = fpr->next)
    {
	siptr = new_instr(s_op_std);
	siptr->src1 = reg_operand(fpr->num + 32);
	siptr->dst = sp_offset_operand((fp_save_offset + ((fpr->num - 6) * 8)));
    }

// Set up for returning struct
    if ((instr->type->base == RECORD_TYPE))
    if (instr->u.call.dst)
    {
	assert (ret_record_index < 256);
	siptr = new_instr(s_op_add);
	siptr->src1 = reg_operand(30);
	siptr->src2 = immed_operand(ret_record_offset[ret_record_index]);
	siptr->dst = reg_operand(1);
	siptr = new_instr(s_op_st);
	siptr->src1 = reg_operand(1);
	siptr->dst = sp_offset_operand(64);

	assert (instr->u.call.dst->kind == TEMP_REG);
	tptr = newTEMP(instr->u.call.dst->num);
	tptr->type = 4;
	tptr->fp = 0;
	tptr->immed = ret_record_offset[ret_record_index];
	ret_record_index++;
    }

// Get target (label or register) and make call
    target = CallTarget(instr->u.call.proc, &sym);
    if (target == -1)
    {
	//printf("!sym = %d\n", sym); fflush(stdout);

	siptr = new_instr(s_op_call); 
	siptr->src1 = call_target(sym);
	siptr->src2 = immed_operand(0);
    }
    else
    {
	siptr = new_instr(s_op_call); 
	siptr->src1 = reg_operand(target);
	siptr->src2 = immed_operand(0);
    }
    new_instr(s_op_nop);
    if ((instr->type->base == RECORD_TYPE))
    if (instr->u.call.dst)
    {
	siptr = new_instr(s_op_unimp); 
	siptr->src1 = immed_operand(instr->u.call.dst->var->type->len >> 3);
    }
// Restore fp registers
    for (fpr = fpreg.reglist(); fpr; fpr = fpr->next)
    {
	siptr = new_instr(s_op_ldd);
	siptr->src1 = sp_offset_operand(fp_save_offset + ((fpr->num - 6) * 8));
	siptr->dst = reg_operand(fpr->num + 32);
    }
// HANDLE RETURNED VALUE
    if ((instr->type->base == SIGNED_TYPE) |
        (instr->type->base == UNSIGNED_TYPE) |
        (instr->type->base == ADDRESS_TYPE))
    if (instr->u.call.dst)
    {
        dst = getDST(0,instr->u.call.dst);
        siptr = new_instr(s_op_mov);
	siptr->src1 = reg_operand(8);
        siptr->dst = reg_operand( dst); 
    }
    if ((instr->type->base == FLOAT_TYPE))
    if (instr->u.call.dst)
    {
	dst = getDST(0,instr->u.call.dst);
	siptr = new_instr(s_op_fmovs);
	siptr->src1 = reg_operand(32);
	siptr->dst = reg_operand( dst); 
	if (instr->type->len == 64);
	{
	    siptr = new_instr(s_op_fmovs);
	    siptr->src1 = reg_operand(33);
	    siptr->dst = reg_operand( dst+1); 
	}
    }
// AND THAT IS ALL
    fflush(stdout);
    return;
}

//********************
void handle_mbr(simple_instr *instr,char *proc_name)
// translate MBR_FORM (multi-way branch) instructins
{
    int i;
    int src;
    s_instr *siptr;
    char temp_str[80];

    src = RegSRC(instr->u.mbr.src,2);
    siptr = new_instr(s_op_sub); 
    siptr->src1 = src_operand( src); 
    siptr->src2 = immed_operand(instr->u.mbr.offset);
    siptr->dst = reg_operand(4);

    // check lower bound
    siptr = new_instr(s_op_cmp);
    siptr->src1 = reg_operand(4);
    siptr->src2 = zero_operand();
    siptr = new_instr(s_op_bl);
    siptr->src1 = jump_target(instr->u.mbr.deflab->name,proc_name);

    // check upper bound
    siptr = new_instr(s_op_cmp);
    siptr->src1 = reg_operand(4);
    siptr->src2 = immed_operand(instr->u.mbr.ntargets);
    siptr = new_instr(s_op_bge);
    siptr->src1 = jump_target(instr->u.mbr.deflab->name,proc_name);

    siptr = new_instr(s_op_sll);
    siptr->src1 = reg_operand(4);
    siptr->src2 = immed_operand(2);
    siptr->dst = reg_operand(4);

    siptr = new_instr(s_op_sethi);
    sprintf(temp_str,"%s_%d", proc_name, quinn+1);
    siptr->src1 = jump_target("_quinn", temp_str);
    siptr->src1->type = 06;
    siptr->dst = reg_operand(1);
    siptr = new_instr(s_op_or);
    siptr->src1 = reg_operand(1);
    sprintf(temp_str,"%s_%d", proc_name, quinn+1);
    siptr->src2 = jump_target("_quinn", temp_str);
    siptr->src2->type = 03;
    siptr->dst = reg_operand(1);

    siptr = new_instr(s_op_add);
    siptr->src1 = reg_operand(4);
    siptr->src2 = reg_operand(1);
    siptr->dst = reg_operand(4);

    siptr = new_instr(s_op_label);
    sprintf(temp_str,"%s_%d", proc_name, quinn);
    siptr->src1 = quinn_label("_quinn", temp_str);
    quinn++;

    siptr = new_instr(s_op_jmp);
    siptr->src1 = new_operand(); 
    siptr->src1->type = 13;
    siptr->src1->immed = 0;
    siptr->src1->reg_num = 4;

    new_instr(s_op_nop);

    siptr = new_instr(s_op_label);
    sprintf(temp_str,"%s_%d", proc_name, quinn);
    siptr->src1 = quinn_label("_quinn", temp_str);
    quinn++;

    for (i = 0; unsigned(i) < instr->u.mbr.ntargets; i++)
    {
	siptr = new_instr(s_op_b_a);
	siptr->src1 = jump_target(instr->u.mbr.targets[i]->name,proc_name);
    }
    siptr = new_instr(s_op_b_a);
    siptr->src1 = jump_target(instr->u.mbr.deflab->name,proc_name);
    return;
}

void handle_label(simple_instr* instr_in, char *proc_name)
// translate LABEL_FORM (label place holders in the code) instructins
{
    simple_instr *instr;
    s_instr	*siptr;

    instr = instr_in;
    siptr = new_instr(s_op_label);
    siptr->src1 = label_string(instr->u.label.lab->name,proc_name);
    return;
}

/************************************************************************
 * Procedure called from outside
************************************************************************/

void proc_instructions(simple_instr* inlist, char* proc_name)
{
    simple_instr *instr; 
    simple_instr *begining;

    // assert(inlist->prev);
    // assert(inlist->prev->opcode == LABEL_OP);
    // begining = inlist->prev;
    begining = inlist;
    TEMPreset();
    for (instr = begining; instr; instr = instr->next)
    {
	mcpy_size = 0;
	//printf("! %s\n", simple_op_name(instr->opcode)); fflush(stdout);
	switch (simple_op_format(instr->opcode)) 
	{
	    case BASE_FORM :
		handle_base(instr,proc_name);
		break;
	    case BJ_FORM :
		handle_bj(instr,proc_name);
		break;
	    case LDC_FORM :
		handle_ldc(instr);
		break;
	    case CALL_FORM :
		handle_call(instr);
		break;
	    case MBR_FORM :
		handle_mbr(instr,proc_name);
		break;
	    case LABEL_FORM :
		handle_label(instr,proc_name);
		break;
	}
    }

}
