////////////////////////////////////////////////////////////////////////
//
//      Quinn Jacobson (qjacobso@ece.wisc.edu)
//      ECE department U of Wisconsin
//
//      CS 701
//      Feb 1996
//
//      reg_alloc.h
//
////////////////////////////////////////////////////////////////////////
//
// Definitions for register allocation needed by multiple files.
//
////////////////////////////////////////////////////////////////////////

#include "simple.h"

#define big_number 0x8000 
// Added to pseudo register number for variables that do not cross function
// calls. This tags that for these operands %o registers can be used for temps.

/* BASIC BLOCK:
   Comprises linked lists of 
   - predecessors
   - successors
   - component instructions
     - should belong to only one basic block
     - id end of chain by match of "last"
   Ptr to another basic block so can easily chain together a set
   (for easy search and free)
   Label is unique identifier (may be NULL)
*/
struct basic_block_node_struct;

typedef struct basic_block_struct {
  struct basic_block_node_struct* pred_head;
  struct basic_block_node_struct* pred_tail;

  struct basic_block_node_struct* succ_head;
  struct basic_block_node_struct* succ_tail;

  simple_instr* instr_head;
  simple_instr* instr_tail;
} basic_block;

typedef struct basic_block_node_struct {
  basic_block* block;
  basic_block_node_struct* next;
} basic_block_node;

void debug_instr_types(simple_instr* inlist);
basic_block_node* build_cfg(simple_instr* inlist);
basic_block* create_block();
basic_block_node* create_node();
void add_instr(basic_block* block, simple_instr* instr);
void add_block_pred(basic_block* cur, basic_block* pred);
void add_block_succ(basic_block* cur, basic_block* succ);
basic_block* find_target_block(basic_block_node* block_list, 
                               simple_sym* target);
void print_cfg(basic_block_node* cfg);
void print_instr(simple_instr* instr);
void delete_cfg(basic_block_node* cfg);
void delete_block(basic_block* block);
