////////////////////////////////////////////////////////////////////////
//
//      Quinn Jacobson (qjacobso@ece.wisc.edu)
//      ECE department U of Wisconsin
//
//      CS 701 
//      Feb 1996
//
//      reg_class.h
//
////////////////////////////////////////////////////////////////////////
//
//    typedef struct REGnode
//    {
//        int         num;
//        struct REGnode *next;
//        struct REGnode *prev;
//    } REGnode;
//
//    REGnode* reglist();
//	returns a pointer to the head of the list. This is used for fp saves/
//	restores accross calls.
//
//    REGnode* getreg(TEMPnode *temp);
//	return a pointer to an available register. If none available: SPILL
//	Needs to know which temp it is associated with so that temp can be 
//  	be updated if the reg it thinks it has has been spilled.
//
//    void freereg(REGnode* reg);
//	free up a register that is in use.
//
//    void REGinit(int is_fp, int temp_offset)
//	Initialize a register class. Sets it to having no registers.
//	Identify if it is of fp regs or not.
//	Temp_offset is the offset from the %fp of the register classes spill
//	space. The spill space is of fixed size of 16 registers.
//
//    void REGreset();
//	Reset registers to all being free, used at begining of a procedure.
//
//    void REGadd(int reg_num);
//	Add a register of number reg_num to the list of registers that are
//	allocatable
//
//    int REGreg(REGnode *reg, int def);
//	Return the register number associated with a register. def is a 
// 	default register that can be used if the register has been spilled
//	and needs to be reloaded for use. If it is a destination then register
//	must be an allocated physical register and def should be 0.
//
////////////////////////////////////////////////////////////////////////
// THAT IS ALL YOU NEED TO KNOW
////////////////////////////////////////////////////////////////////////


/************************************************************************
 * STRUCTURE
************************************************************************/
#define MAX_OVERFLOW 	16
#define	OVER_REG	0x8000
#define MAX_REG		32

extern int 	special_temp;

typedef struct REGnode
{
    int         num;
    int		offset;
    int		is_load;
    int		bad;
    struct TEMPnode *temp;
    struct REGnode *next;
    struct REGnode *prev;
} REGnode;

/************************************************************************
 * CLASS
************************************************************************/
class reg_class
{

private:

    REGnode *free_list;
    REGnode *alloc_list;
    struct TEMPnode *temp;
    int	    temp_storage_offset;
    char    temp_storage[MAX_OVERFLOW];
    int	    is_fp_regs;

public:

    //*******************
    REGnode* reglist()
    {
	return (alloc_list);
    }

    //*******************
    REGnode* getreg(TEMPnode *temp, int is_load)
    {
        REGnode *reg;
	REGnode *new_reg;
	int	i;
	int	offset;
	s_instr *siptr;
    

        reg = free_list;
	if ((special_temp) && (is_load))
	{
	    if(reg)
	    {
	        while (reg->next) 
 		    reg = reg->next;
	    }
	}

        if (reg == NULL)
        {
	    printf("! SPILL\n");
	    // find reg to reallocate
	    for (reg = alloc_list; reg->next; reg = reg->next) {};
	    // find empty slot of temp storage space
	    for (i = 0; i < MAX_OVERFLOW; i++)
	    {
		if (temp_storage[i] == 0)
		    break;
	    }
	    // if no temp storage space die
	    if (i == MAX_OVERFLOW)
	    {
		fprintf(stderr,"BAD NEWS: ran out of temp overflow space!\n");
		exit (1);
	    }

	    // Save value previously in reg
	    temp_storage[i] = 1;
	    if (is_fp_regs)
	    {
		offset = temp_storage_offset + (i*8);
		siptr = new_instr(s_op_std);
		siptr->src1 = reg_operand(reg->num + 32);
		siptr->dst = fp_offset_operand(offset);
	    }
	    else
	    {
		offset = temp_storage_offset + (i * 4);
		siptr = new_instr(s_op_st);
		siptr->src1 = reg_operand(reg->num);
		siptr->dst = fp_offset_operand(offset);
	    }
	    offset += OVER_REG;
	    assert (offset > MAX_REG);
	    // make dummy register to correspond to saved value
	    new_reg = (REGnode *)malloc(sizeof(REGnode));
	    assert (new_reg);
	    new_reg->bad = 1;
	    new_reg->offset = i;
	    reg->temp->reg = new_reg;
	    // return reallocated reg
	    reg->temp = temp;
	    return (reg);
        }
	if (free_list == reg)
            free_list = free_list->next;
	if (reg->next) reg->next->prev = reg->prev;
	if (reg->prev) reg->prev->next = reg->next;
    
        // Add to alloc list
        reg->prev = NULL;
        reg->next = alloc_list;
	reg->temp = temp;
        if (alloc_list) alloc_list->prev = reg;
        alloc_list = reg;

	if ((is_load) & (special_temp))
	    reg->is_load = 1;
   	else
	    reg->is_load = 0; 

        return (reg);
    }

    //*******************
    void freereg(REGnode* reg)
    {
	REGnode* ptr;
	int	cnt;
	int	i;

	if (reg->bad)
	{
	    // Its a dummy node from an overflow
	    temp_storage[reg->offset] = 0;
	    free (reg);
	    return;
	}
        // Remove from alloc list
        if (reg->prev)
	    reg->prev->next = reg->next;
        else
	    alloc_list = reg->next;
        if (reg->next)
	    reg->next->prev = reg->prev;
     
        // Add to free list
	if (reg->is_load)
	{
	    cnt = 0;
	    for (ptr = free_list; ptr; ptr = ptr->next)
		cnt++;
	    ptr = free_list;
	    if (ptr)
	        for (i = 0; i < (cnt-1)/2; i++)
		    if (ptr->next) ptr = ptr->next;
	    if (ptr)
	    {
		reg->prev = ptr;
		reg->next = ptr->next;
		if (reg->next) reg->next->prev = reg;
		ptr->next = reg;
	    }
	    else
	    {

            	reg->prev = NULL;
            	reg->next = free_list;
	    	reg->temp = NULL;
            	if (free_list) free_list->prev = reg;
            	free_list = reg;
	    }
	}
	else
	{	
            reg->prev = NULL;
            reg->next = free_list;
	    reg->temp = NULL;
            if (free_list) free_list->prev = reg;
            free_list = reg;
	}
    
    }

    //*******************
    void REGinit(int is_fp, int temp_offset)
    {
	int	i;
	REGnode *ptr;
	REGnode *temp;
	
	ptr = free_list;
	while (ptr)
	{
	    temp = ptr;
	    ptr = ptr->next;
	    free (temp);
	}

	ptr = alloc_list;
	while (ptr)
	{
	    temp = ptr;
	    ptr = ptr->next;
	    free (temp);
	}

        free_list = NULL;
        alloc_list = NULL;
	temp_storage_offset = temp_offset;
	is_fp_regs = is_fp;
	for (i = 0; i < MAX_OVERFLOW; i++)
	{
	    temp_storage[i] = 0;
	}
	assert (alloc_list == NULL);
    }
    
    //*******************
    void REGreset()
    {
        REGnode *p;
        REGnode *b;
        b = NULL;
        for (p = alloc_list; p; p = p->next)
        {
            if (b) freereg(b);
            b = p;
        }
        if (b) freereg(b);
	assert (alloc_list == NULL);
    }

    //*******************
    void REGadd(int reg_num)
    {
        REGnode *p;
    
        assert(reg_num >= 0);
        p = (REGnode *)malloc(sizeof(REGnode));
        p->num = reg_num;
	p->offset = 0;
	p->bad = 0;
        p->prev = NULL;
        p->next = free_list;
        if (free_list) free_list->prev = p;
        free_list = p;
    }

    //*******************
    int	REGreg(REGnode *reg, int def)
    {
	int	offset;
	s_instr *siptr;

	if (reg->bad)
	{
	    printf("! UNSPILL\n");
	    if (def == 0)
		assert(0);

	    if (is_fp_regs)
	    {
		offset = temp_storage_offset + (reg->offset * 8);
		siptr = new_instr(s_op_ldd);
		siptr->src1 = fp_offset_operand(offset);
		siptr->dst = reg_operand(def + 32);
	    }
	    else
	    {
		offset = temp_storage_offset + (reg->offset * 4);
		siptr = new_instr(s_op_ld);
		siptr->src1 = fp_offset_operand(offset);
		siptr->dst = reg_operand(def);
	    }
	    return(def);
	}
	else
	{
	    return (reg->num);
	}
    }
};
