////////////////////////////////////////////////////////////////////////
//
//      Quinn Jacobson (qjacobso@ece.wisc.edu)
//      ECE department U of Wisconsin
//
//      CS 701
//      March 1996
//
//      reg_pass.cc
//
////////////////////////////////////////////////////////////////////////
//
// Performs a pass of the suif instructions to manipulate register operands
// Updates the suif:
//      for all TEMP_REGs determine if their life span crosses a call or not.
//	Identify regs that do NOT cross call by adding big_number to their
//	temp number. 
//
////////////////////////////////////////////////////////////////////////

//----- Include -----
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <assert.h>
 
extern "C" {
#include <simple.h>
}
 
#include "reg_alloc.h"
#include "ss_symtab.h"
#include "LocalVars.h"
#include "reg_pass.h"

//----- CODE -----

//***************
//** TEMP CLASSIFICATION
//***************

simple_reg* check_src(int num, simple_reg *reg)
{
    if (reg == NULL)
	return (NULL);

    if ((reg->kind == TEMP_REG) && (reg->num == num))
    {
	return (reg);
    }

    return (NULL);
}

//***************
simple_reg* find_src(int num, simple_instr *first)
{
    simple_instr   *instr;
    simple_reg	   *reg;

    for (instr = first->next; instr; instr = instr->next)
    {
	switch (simple_op_format(instr->opcode))
	{
	    case BASE_FORM :
		if (instr->opcode == REM_OP) return (NULL);
		if (instr->opcode == MOD_OP) return (NULL);
		// BUG FIX:
		//   an integer divide causes a call to .div to be
		//   generated, so treat a DIV_OP as a call if the
		//   type is not FLOAT_TYPE
		if (instr->opcode == DIV_OP &&
		    instr->type->base != FLOAT_TYPE) return(NULL);
		reg = check_src(num, instr->u.base.src1);
		if (reg) return (reg);
		reg = check_src(num, instr->u.base.src2);
		if (reg) return (reg);
		break;
	    case BJ_FORM :
		reg = check_src(num, instr->u.bj.src);
		if (reg) return (reg);
		return (NULL);
		break;
	    case LDC_FORM :
		break;
	    case CALL_FORM :
		return (NULL);
		break;
	    case MBR_FORM :
		reg = check_src(num, instr->u.mbr.src);
		if (reg) return (reg);
		return (NULL);
		break;
	    case LABEL_FORM :
		return (NULL);
		break;
	}
    }
    return (NULL);
}

//***************
void temp_dst(simple_reg *reg, simple_instr *instr)
{
    int		num;
    simple_reg	*src;

    num = reg->num;
    assert(num < big_number);
    src = find_src(num, instr);

    if (instr->opcode == CVT_OP)
    if (instr->u.base.src1->var->type->base != FLOAT_TYPE)
    if (instr->u.base.dst->var->type->base != FLOAT_TYPE)
    if (instr->u.base.src1->var->type->len == 32)
    if (instr->u.base.dst->var->type->len == 32)
    {
	// Useless convert
        if (instr->u.base.src1->kind == TEMP_REG)
        {
	    // If source and dst regs of different TEMP class make both cross
            if (instr->u.base.src1->num < big_number)
	        src = NULL;
	    else if (src == NULL)
	    {
	        instr->u.base.src1->num -= big_number;
	    }
        }
	if (instr->u.base.src1->kind == MACHINE_REG)
	{
	    // If source is MACHINE_REG propergate it down to next use
	    // and remove useless instruction
	    reg->kind = MACHINE_REG;
	    reg->num = instr->u.base.src1->num;
	    reg->var = instr->u.base.src1->var;
	    if (instr->prev)
	        instr->prev->next = instr->next;
	    if (instr->next)
	        instr->next->prev = instr->prev;
	    return;
	}
    }
    if (src)
    {
	reg->num = num + big_number;
	assert (src->num == num + big_number);
	//src->num = num + big_number;
    }
    return; 
}


//***************
//** disect an instruction
//***************

void reg_operand(simple_reg **reg, int is_src, simple_instr *instr)
{
    if (*reg == NULL)
    {
	return;
    }
    else if (((*reg)->kind == TEMP_REG) && (is_src == 0))
    {
	temp_dst(*reg, instr);
    }
}

//***************
void look_at_instr(simple_instr	*instr)
{
    int		i;

    switch (simple_op_format(instr->opcode))
    {
	case BASE_FORM :
	    reg_operand(&(instr->u.base.src1), 1, instr);
	    reg_operand(&(instr->u.base.src2), 1, instr);
	    reg_operand(&(instr->u.base.dst), 0, instr);
	    break;
	case BJ_FORM :
	    reg_operand(&(instr->u.bj.src), 1, instr);
	    break;
	case LDC_FORM :
	    reg_operand(&(instr->u.ldc.dst), 0, instr);
	    break;
	case CALL_FORM :
	    for (i = 0; unsigned(i) < instr->u.call.nargs; i++)
	    {
		reg_operand(&(instr->u.call.args[i]), 1, instr);
	    }
	    reg_operand(&(instr->u.call.dst), 0, instr);
	    break;
	case MBR_FORM :
	    reg_operand(&(instr->u.mbr.src), 1, instr);
	    break;
	case LABEL_FORM :
	    break;
    }
    return;
}

//***************
//** External Access
//***************

void reg_pass(simple_instr* inlist)
{
    simple_instr 	*instr;
    
    for (instr=inlist; instr; instr=instr->next)
    {
        look_at_instr(instr);
    }
}
