/************************************************************************
 * hash_utils.h: hashing utilities.
************************************************************************/





#if !defined(_hash_utils_h_)
#define _hash_utils_h_





/************************************************************************
 * types.
************************************************************************/

typedef unsigned long hash_t;





/************************************************************************
 * namespace STRINGS
 *
 * hash_t string_hash(const char* first, const char* last)
 * hash_t string_hash(const string& str)
 *
 * hash `str' (possibly from `first' to `last') into a hash_t.
************************************************************************/

namespace STRINGS {

#include <string>

    hash_t internal_string_hash(const char* first, const char* last) {
        hash_t hv = 5381;
        for (; first != last; ++first) {
            hv = 33*hv + (hash_t)(*first);
        }
        return hv;
    }

    hash_t string_hash(const string& str) {
        return internal_string_hash(str.begin(), str.end());
    }
};





/************************************************************************
 * template<class T> class hash
************************************************************************/

template<class T>
class hash {
private:
    typedef hash_t (*hash_function)(const T &);
    hash_function hf_;

public:
    hash(hash_function hf) : hf_(hf) {
    }

    hash(const hash<T>& h) : hf_(h.hf_) {
    }

    const hash<T>& operator=(const hash<T>& h) {
        if (this != &h) {
            hf_ = h.hf_;
        }
        return *this;
    }

    bool operator==(const hash<T>& h) const {
        return ((this == &h) || (hf_ == h.hf_));
    }

    hash_t operator()(const T& t) const {
        return hf_(t);
    }
};





#endif /* !defined(_hash_utils_h_) */
