////////////////////////////////////////////////////////////////////////
//
//      Quinn Jacobson (qjacobso@ece.wisc.edu)
//      ECE department U of Wisconsin
//
//      CS 701
//      Feb 1996
//
//      temp.cc
//
////////////////////////////////////////////////////////////////////////
//
// Temporary mapping
//
////////////////////////////////////////////////////////////////////////

//----- Includes -----
 
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "reg_alloc.h"
#include "temp.h"
#include "instr.h"
#include "reg_class.h"

#define TEMPhash 1024

//----- GLOBALS -----

extern reg_class inreg;
extern reg_class ccreg;
extern reg_class fpreg;

TEMPnode *TEMPtable[TEMPhash];

//----- Code -----

//***************
void TEMPinit()
// Initialize data structure
{
    int i;

    inreg.REGinit(0, -144);
    ccreg.REGinit(0, -80);
    fpreg.REGinit(1, -272);
    for (i = 0; i < TEMPhash; i++)
        TEMPtable[i] = NULL;
}

//***************
void TEMPreset()
// Reset data structures at the beginning of processing a function
{
        int i;
        TEMPnode *t;
        TEMPnode *bt;
   
   
	inreg.REGreset();
	fpreg.REGreset();
        for (i = 0; i < TEMPhash; i++)
        {
            bt = NULL;
            for (t = TEMPtable[i]; t; t = t->next)
            {
                if (bt) free(bt);
                bt = t;
            }
            if (bt) free(bt);
            TEMPtable[i] = NULL;
        }
}

//***************
TEMPnode* findTEMP(int temp)
// Return a pointer to the TEMPnode associated with TEMP_REG # temp
{
    int 	i;
    TEMPnode *tptr;

    i = temp % TEMPhash;
    for (tptr = TEMPtable[i]; tptr; tptr = tptr->next)
	if (tptr->temp == temp) return (tptr);
    fprintf(stderr,"-- ERROR -- findreg(): Temp not valid!\n");
    exit (1);
}

//***************
void removeTEMP(TEMPnode *tptr)
// Remove a TEMPnode that has been used
{
    int 	i;

    assert(tptr);
    i = tptr->temp % TEMPhash;
    if (tptr->reg) 
    {
	if (tptr->fp) 
	    fpreg.freereg(tptr->reg);
	else if (tptr->temp >= big_number) 
	    inreg.freereg(tptr->reg);
	else 
	    ccreg.freereg(tptr->reg);
    }
    if (tptr->next) tptr->next->prev = tptr->prev;
    if (tptr->prev) tptr->prev->next = tptr->next;
    else            TEMPtable[i] = tptr->next;
    free(tptr);
}

//***************
TEMPnode* newTEMP(int temp)
// allocate a new TEMPnode to be associated with the TEMP_REG # temp
{
    int		i;
    TEMPnode *tptr;

    i = temp % TEMPhash;
    tptr = (TEMPnode *)malloc(sizeof(TEMPnode));
    tptr->temp = temp;
    tptr->next = TEMPtable[i];
    tptr->prev = NULL;
    tptr->reg = NULL;
    tptr->fp = 0;
    tptr->size = 0;
    if (TEMPtable[i]) TEMPtable[i]->prev = tptr;
    TEMPtable[i] = tptr;

    return (tptr);
}

